# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PyTreeCheckpointHandler class.

Implementation of `CheckpointHandler` interface dealing with JAX PyTrees. Much
of the underlying reading/writing logic for individual leaf types can be
customized, and is delegated to the `TypeHandler` class.
"""

from __future__ import annotations

import asyncio
import dataclasses
import json
import re
import typing
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from absl import logging
from etils import epath
import jax
import numpy as np
from orbax.checkpoint import aggregate_handlers
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint import transform_utils
from orbax.checkpoint import utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.handlers import async_checkpoint_handler
from orbax.checkpoint._src.handlers import base_pytree_checkpoint_handler
from orbax.checkpoint._src.metadata import array_metadata_store as array_metadata_store_lib
from orbax.checkpoint._src.metadata import empty_values
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint._src.tree import types as tree_types
from orbax.checkpoint._src.tree import utils as tree_utils
import tensorstore as ts


PyTree = Any
TupleKey = Tuple[str, ...]
RestoreArgs = type_handlers.RestoreArgs
ArrayRestoreArgs = type_handlers.ArrayRestoreArgs
SaveArgs = type_handlers.SaveArgs
ParamInfo = type_handlers.ParamInfo
TypeHandler = type_handlers.TypeHandler
TypeHandlerRegistry = type_handlers.TypeHandlerRegistry
AggregateHandler = aggregate_handlers.AggregateHandler
MsgpackHandler = aggregate_handlers.MsgpackHandler
LegacyTransformFn = Callable[[PyTree, PyTree, PyTree], Tuple[PyTree, PyTree]]
Transform = transform_utils.Transform
RestoreTransform = transform_utils.RestoreTransform
CheckpointArgs = checkpoint_args.CheckpointArgs
register_with_handler = checkpoint_args.register_with_handler

BasePyTreeCheckpointHandler = (
    base_pytree_checkpoint_handler.BasePyTreeCheckpointHandler
)
BasePyTreeSaveArgs = base_pytree_checkpoint_handler.BasePyTreeSaveArgs
BasePyTreeRestoreArgs = base_pytree_checkpoint_handler.BasePyTreeRestoreArgs
LimitInFlightBytes = base_pytree_checkpoint_handler.LimitInFlightBytes
get_param_names = base_pytree_checkpoint_handler.get_param_names

PYTREE_METADATA_FILE = base_pytree_checkpoint_handler.PYTREE_METADATA_FILE
_CHECKPOINT_FILE = 'checkpoint'
_METADATA_FILE = PYTREE_METADATA_FILE
DEFAULT_CONCURRENT_GB = base_pytree_checkpoint_handler.DEFAULT_CONCURRENT_GB


def _maybe_set_default_restore_args(args):
  if isinstance(args, RestoreArgs):
    return args
  return RestoreArgs(restore_type=None)


def _try_array_cast(arr, dtype):
  if dtype is not None:
    if utils.is_scalar(arr):
      arr = np.asarray(arr).astype(dtype).item()
    else:
      if hasattr(arr, 'astype'):
        arr = arr.astype(dtype)
  return arr


def _maybe_shard_array(value, args):
  if hasattr(value, 'reshape') and isinstance(args, ArrayRestoreArgs):
    value = value.reshape(args.global_shape)
    sharding = args.sharding or jax.sharding.NamedSharding(
        args.mesh, args.mesh_axes
    )
    value = jax.make_array_from_callback(
        value.shape, sharding, lambda idx: value[idx]
    )
  return value


def _keystr(key: Tuple[Any, ...]) -> str:
  return '/'.join(key)


def _find_matching_input_args(
    input_key: TupleKey,
    flat_item: Dict[TupleKey, Any],
    flat_transforms: Dict[TupleKey, Transform],
    flat_restore_args: Dict[TupleKey, RestoreArgs],
) -> Optional[RestoreArgs]:
  """Given an input_key, tries to find matching RestoreArgs for the input.

  Args:
    input_key: A key in the input tree.
    flat_item: The flattened, user-provided item.
    flat_transforms: Flattened transformations dict.
    flat_restore_args: Flattened tree of RestoreArgs, relative to item.

  Returns:
    RestoreArgs that match the given input_key, according to the
    transformations, or None if no match is found.
  """
  for transform_key, transform in flat_transforms.items():
    if transform.multi_value_fn is not None:
      if not isinstance(transform, RestoreTransform):
        raise ValueError(
            'Must use RestoreTransform in order to use multi_value_fn'
            ' during restore.'
        )
      if transform.multi_value_fn_input_args is None:
        raise ValueError(
            '`multi_value_fn` was specified, but'
            ' `multi_value_fn_input_args` were not. The latter must be'
            ' specified to identify inputs for the function.'
        )
      for (
          input_key_regex,
          input_args,
      ) in transform.multi_value_fn_input_args.items():
        if re.fullmatch(input_key_regex, _keystr(input_key)):
          return input_args
    elif not transform.use_fallback:
      # The following is done to reverse-engineer the regex for the key in
      # the original tree.
      for output_key in flat_item:
        match = re.fullmatch(_keystr(transform_key), _keystr(output_key))
        if match:
          if transform.original_key is None:
            # If transform.original_key is not specified, this transform
            # does not rename the original key. We can reuse the key from
            # the item.
            input_key_pattern = _keystr(output_key)
          else:
            input_key_pattern = match.expand(transform.original_key)
          if input_key_pattern == _keystr(input_key):
            return flat_restore_args[output_key]
  return None


def _has_use_fallback_transform(
    input_key: TupleKey, flat_transforms: Dict[TupleKey, Transform]
) -> bool:
  result = False
  for transform_key, transform in flat_transforms.items():
    match = re.fullmatch(_keystr(transform_key), _keystr(input_key))
    if match and transform.use_fallback:
      result = True
  return result


def _get_restore_parameters(
    directory: epath.Path,
    item: Optional[PyTree],
    structure: PyTree,
    param_names: Optional[PyTree],
    transforms: Optional[PyTree],
    restore_args: Optional[PyTree],
    pytree_metadata_options: tree_metadata.PyTreeMetadataOptions,
    byte_limiter: Optional[LimitInFlightBytes] = None,
    transforms_default_to_original: bool = True,
    use_zarr3: bool = False,
) -> Tuple[PyTree, PyTree]:
  """Construct parameters needed for restoration.

  If transforms are not provided, the method is pretty simple: param_infos are
  constructed from the structure of the original checkpoint, and restore_args
  are serialized to a tree structure compatible with param_infos and structure.

  If transforms are provided, things become more complicated because we must
  determine exactly which parameters the user desires to restore, and construct
  param_infos and restore_args for these, while discarding unneeded parameters.
  In essence, the process can be thought of as reversing the transformations.
  This happens differently for different types of transforms.
  1. Renamed key: Identify the original key name (in the checkpoint) and carry
    over the provided restore args for the parameter.
  2. multi_value_fn: Users are required to specify multi_value_fn_input_args.
    Any keys named here must be loaded, and their restore args are also given
    here.
  3. Unspecified key: A key which is unspecified in the transforms but present
    in the `item` is a key that is carried over from the checkpoint unchanged.
  4. Fallback key: This is a key that is present in the `item` but not in the
    original checkpoint. It does not need to be restored.
  5. Keys present in the original checkpoint but not in the `item`/`transforms`
    are implicitly ignored, and not restored.

  Args:
    directory: Checkpoint directory.
    item: Optional reference item.
    structure: The structure of the original checkpoint.
    param_names: Tree of parameter names.
    transforms: User-provided transformations. If None, they were not provided.
      Has the structure of the desired output tree.
    restore_args: User-provided restoration arguments. If None, they were not
      provided. Otherwise, the tree has the same structure as the desired output
      tree.
    pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
    byte_limiter: A LimitInFlightBytes object.
    transforms_default_to_original: See transform_utils.apply_transformations.
    use_zarr3: If True, use Zarr ver3 otherwise Zarr ver2

  Returns:
    Tuple of param_infos, and restore_args.
  """
  flat_structure = tree_utils.to_flat_dict(structure, keep_empty_nodes=True)
  if param_names is None:
    param_names = get_param_names(structure)
  flat_param_names = tree_utils.to_flat_dict(param_names, keep_empty_nodes=True)
  if restore_args is None:
    restore_args = jax.tree.map(lambda x: RestoreArgs(), structure)
  flat_restore_args = tree_utils.to_flat_dict(
      restore_args, keep_empty_nodes=True
  )
  flat_param_infos = {}
  flat_input_restore_args = {}
  is_ocdbt_checkpoint = type_handlers.is_ocdbt_checkpoint(directory)
  ts_context = ts_utils.get_ts_context(use_ocdbt=is_ocdbt_checkpoint)

  def _get_param_info(
      name: str,
      meta_or_value: Union[Any, tree_metadata.ValueMetadataEntry],
  ) -> Union[ParamInfo, Any]:
    if empty_values.is_supported_empty_value(
        meta_or_value, pytree_metadata_options
    ):
      # Empty node, ParamInfo should not be returned.
      return meta_or_value
    elif not isinstance(meta_or_value, tree_metadata.ValueMetadataEntry):
      # Aggregated value.
      skip_deserialize = True
    else:
      skip_deserialize = meta_or_value.skip_deserialize
    return ParamInfo(
        name=name,
        path=directory / name,
        parent_dir=directory,
        skip_deserialize=skip_deserialize,
        is_ocdbt_checkpoint=is_ocdbt_checkpoint,
        byte_limiter=byte_limiter,
        use_zarr3=use_zarr3,
        ts_context=ts_context,
        # Skip raising array data missing error on this code path, since it
        # almost exclusively handles legacy use cases.
        raise_array_data_missing_error=False,
    )

  if transforms is None:
    for key, meta in flat_structure.items():
      flat_param_infos[key] = _get_param_info(flat_param_names[key], meta)
    restore_args = tree_metadata.serialize_tree(
        restore_args, pytree_metadata_options
    )
  else:
    if item is None:
      raise ValueError(
          'If providing `transforms`, must provide `item` matching structure'
          ' of expected result.'
      )
    flat_item = tree_utils.to_flat_dict(item, keep_empty_nodes=True)
    flat_transforms = tree_utils.to_flat_dict(transforms)

    for input_key, meta in flat_structure.items():
      maybe_input_args = _find_matching_input_args(
          input_key, flat_item, flat_transforms, flat_restore_args
      )
      if maybe_input_args:
        flat_param_infos[input_key] = _get_param_info(
            flat_param_names[input_key], meta
        )
        flat_input_restore_args[input_key] = maybe_input_args
      elif input_key in flat_item and input_key in flat_structure:
        # Key is present in both input and output.
        if _has_use_fallback_transform(input_key, flat_transforms):
          # Indicates that a `use_fallback` transformation was specified.
          if transforms_default_to_original:
            # Specified `use_fallback`, but key was also present in the
            # checkpoint. This means we should skip loading, since it will be
            # overridden with a new value.
            flat_param_infos[input_key] = ParamInfo(skip_deserialize=True)
            flat_input_restore_args[input_key] = RestoreArgs()
          else:
            # Specified `use_fallback`, but `transforms_default_to_original`
            # is False. This means we draw the value from the user-provided
            # `item`.
            flat_param_infos[input_key] = _get_param_info(
                flat_param_names[input_key], meta
            )
            flat_input_restore_args[input_key] = flat_restore_args[input_key]
        else:
          # Transform not specified.
          if transforms_default_to_original:
            # Key/value is carried over from the original unchanged.
            flat_param_infos[input_key] = _get_param_info(
                flat_param_names[input_key], meta
            )
            flat_input_restore_args[input_key] = flat_restore_args[input_key]
          else:
            # Take the value from the user-provided `item`, ignoring any value
            # in the checkpoint.
            flat_param_infos[input_key] = ParamInfo(skip_deserialize=True)
            flat_input_restore_args[input_key] = RestoreArgs()
      else:
        # No match, restoration not required since it will be dropped from the
        # output.
        flat_param_infos[input_key] = ParamInfo(skip_deserialize=True)
        flat_input_restore_args[input_key] = RestoreArgs()

    restore_args = tree_utils.from_flat_dict(
        flat_input_restore_args, target=structure
    )

  return (
      tree_utils.from_flat_dict(flat_param_infos, target=structure),
      restore_args,
  )


def _multi_value_fns_with_args(
    transforms: PyTree, restore_args: PyTree
) -> PyTree:
  """Constructs a wrapper for multi_value_fn including RestoreArgs."""
  flat_restore_args = tree_utils.to_flat_dict(restore_args, sep='/')

  def _maybe_wrap_transform(transform: Transform):
    def _multi_value_fn_with_args(transform_key: str, tree: PyTree) -> Any:
      nonlocal transform
      transform = typing.cast(RestoreTransform, transform)
      return transform.multi_value_fn(
          transform_key, tree, flat_restore_args[transform_key]
      )

    if transform.multi_value_fn is not None:
      return Transform(multi_value_fn=_multi_value_fn_with_args)
    else:
      return transform

  return jax.tree.map(_maybe_wrap_transform, transforms)


def _transform_checkpoint(
    item: PyTree,
    restored: PyTree,
    restore_args: Optional[PyTree],
    transforms: Optional[PyTree],
    transforms_default_to_original: bool,
) -> PyTree:
  """Optionally transforms the restored PyTree to the structure of `item`.

  Args:
    item: a PyTree representing the result structure ("new tree structure").
    restored: a PyTree representing the original tree structure.
    restore_args: tree of RestoreArgs, with the same structure as `item`.
    transforms: provides instructions on how to transform the input trees. See
      transform_utils.
    transforms_default_to_original: See transform_utils.

  Returns:
    A transformed PyTree.
  """
  if item is None:
    if transforms is not None:
      msg = (
          'If providing `transforms`, must provide `item` matching structure'
          ' of expected result.'
      )
      raise ValueError(msg)
    item = restored
  else:
    if transforms is None:
      item = tree_utils.deserialize_tree(restored, item)
    else:
      if restore_args is None:
        raise ValueError(
            'If providing `transforms`, must provide `restore_args` matching'
            ' structure of expected result.'
        )
      transforms = _multi_value_fns_with_args(transforms, restore_args)
      item = transform_utils.apply_transformations(
          restored, transforms, item, transforms_default_to_original
      )
  return item


def _get_impl_save_args(
    item: Optional[PyTree] = None,
    save_args: Optional[PyTreeSaveArgs] = None,
    args: Optional[PyTreeSaveArgs] = None,
) -> BasePyTreeSaveArgs:
  """Construct BasePyTreeSaveArgs."""
  if isinstance(item, CheckpointArgs):
    raise ValueError(
        'Make sure to specify kwarg name `args=` when providing'
        ' `PyTreeSaveArgs`.'
    )
  if args is None:
    args = PyTreeSaveArgs(
        item=item,
        save_args=save_args,
    )
  return BasePyTreeSaveArgs(
      item=args.item,
      save_args=args.save_args,
      ocdbt_target_data_file_size=args.ocdbt_target_data_file_size,
      enable_pinned_host_transfer=args.enable_pinned_host_transfer,
      custom_metadata=args.custom_metadata,
  )


def _concurrent_bytes(concurrent_gb: Optional[int]) -> int:
  if concurrent_gb is None:
    return DEFAULT_CONCURRENT_GB * 10**9
  else:
    return concurrent_gb * 10**9


class PyTreeCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
  """A CheckpointHandler implementation for any PyTree structure.

  See JAX documentation for more information on what consistutes a "PyTree".
  This handler is capable of saving and restoring any leaf object for which a
  `TypeHandler` (see documentation) is registered. By default, `TypeHandler`s
  for standard types like `np.ndarray`, `jax.Array`, Python scalars, and others
  are registered.

  As with all `CheckpointHandler` subclasses, `PyTreeCheckpointHandler` should
  only be used in conjunction with a `Checkpointer` (or subclass). By itself,
  the `CheckpointHandler` is non-atomic.

  Example::

    ckptr = Checkpointer(PyTreeCheckpointHandler())

  # TODO(cpgaffney) Cut down on the protected methods accessed by this class.
  """

  def __init__(
      self,
      aggregate_filename: Optional[str] = None,
      *,
      save_concurrent_gb: Optional[int] = None,
      restore_concurrent_gb: Optional[int] = None,
      use_ocdbt: bool = True,
      use_zarr3: bool = False,
      multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
      type_handler_registry: TypeHandlerRegistry = type_handlers.GLOBAL_TYPE_HANDLER_REGISTRY,
      handler_impl: Optional[BasePyTreeCheckpointHandler] = None,
      pytree_metadata_options: tree_metadata.PyTreeMetadataOptions = (
          tree_metadata.PYTREE_METADATA_OPTIONS
      ),
      array_metadata_validator: array_metadata_store_lib.Validator = (
          array_metadata_store_lib.Validator()
      ),
  ):
    """Creates PyTreeCheckpointHandler.

    Args:
      aggregate_filename: name that the aggregated checkpoint should be saved
        as.
      save_concurrent_gb: max concurrent GB that are allowed for writing. Can
        help to reduce the possibility of OOM's when large checkpoints are
        saved.
      restore_concurrent_gb: max concurrent GB that are allowed for writing. Can
        help to reduce the possibility of OOM's when large checkpoints are
        restored.
      use_ocdbt: enables Tensorstore OCDBT driver. This option allows using a
        different checkpoint format which is faster to read and write, as well
        as more space efficient.
      use_zarr3: If True, use Zarr ver3 otherwise Zarr ver2
      multiprocessing_options: See orbax.checkpoint.options
      type_handler_registry: a type_handlers.TypeHandlerRegistry. If not
        specified, the global type handler registry will be used.
      handler_impl: Allows overriding the internal implementation.
      pytree_metadata_options: `PyTreeMetadataOptions` to manage metadata.
      array_metadata_validator: Validator for ArrayMetadata.
    """
    self._aggregate_handler = MsgpackHandler(
        primary_host=multiprocessing_options.primary_host,
        pytree_metadata_options=pytree_metadata_options,
    )
    if aggregate_filename is None:
      aggregate_filename = _CHECKPOINT_FILE
    self._aggregate_filename = aggregate_filename
    self._use_ocdbt = use_ocdbt
    self._use_zarr3 = use_zarr3
    self._primary_host = multiprocessing_options.primary_host
    self._type_handler_registry = type_handler_registry
    self._save_concurrent_bytes = _concurrent_bytes(save_concurrent_gb)
    self._restore_concurrent_bytes = _concurrent_bytes(restore_concurrent_gb)
    self._handler_impl = handler_impl or BasePyTreeCheckpointHandler(
        save_concurrent_bytes=self._save_concurrent_bytes,
        restore_concurrent_bytes=self._restore_concurrent_bytes,
        use_ocdbt=use_ocdbt,
        use_zarr3=use_zarr3,
        multiprocessing_options=multiprocessing_options,
        type_handler_registry=type_handler_registry,
        pytree_metadata_options=pytree_metadata_options,
        array_metadata_validator=array_metadata_validator,
    )
    self._pytree_metadata_options = pytree_metadata_options

  async def async_save(
      self,
      directory: epath.Path,
      item: Optional[PyTree] = None,
      save_args: Optional[PyTreeSaveArgs] = None,
      args: Optional[PyTreeSaveArgs] = None,
  ) -> Optional[List[future.Future]]:
    """Saves a PyTree to a given directory.

    This operation is compatible with a multi-host, multi-device setting. Tree
    leaf values must be supported by the type_handler_registry given in the
    constructor. Standard supported types include Python scalars, `np.ndarray`,
    `jax.Array`, and strings.

    After saving, all files will be located in "directory/". The exact files
    that are saved depend on the specific combination of options, including
    `use_ocdbt`. A JSON metadata file will be present to store the
    tree structure.

    Example usage::

      ckptr = ocp.Checkpointer(ocp.PyTreeCheckpointHandler())
      item = {
          'layer0': {
              'w': np.ndarray(...),
              'b': np.ndarray(...),
          },
          'layer1': {
              'w': np.ndarray(...),
              'b': np.ndarray(...),
          },
      }
      # Note: save_args may be None if no customization is desired for saved
      # parameters.
      # Otherwise, settings can be used to customize save behavior, e.g.
      # casting.
      save_args = jax.tree.map(lambda x: ocp.SaveArgs(dtype=np.int32), item)
      # Eventually calls through to `async_save`.
      ckptr.save(path, args=ocp.PyTreeSave(item, save_args))

    Args:
      directory: save location directory.
      item: Deprecated, use `args.
      save_args: Deprecated, use `args`.
      args: `PyTreeSaveArgs` (see below).

    Returns:
      A Future that will commit the data to `directory` when awaited. Copying
      the data from its source will be awaited in this function.
    """
    args = _get_impl_save_args(item, save_args, args)
    return await self._handler_impl.async_save(directory, args=args)

  def save(
      self,
      directory: epath.Path,
      item: Optional[PyTree] = None,
      save_args: Optional[PyTreeSaveArgs] = None,
      args: Optional[PyTreeSaveArgs] = None,
  ):
    """Saves the provided item. See async_save."""
    args = _get_impl_save_args(item, save_args, args)
    self._handler_impl.save(directory, args=args)

  async def _maybe_deserialize(
      self,
      item: PyTree,
      metadata: PyTree,
      param_infos: PyTree,
      restore_args: PyTree,
  ) -> PyTree:
    """Deserializes values or gets them from the aggregate file."""
    byte_limiter = serialization.get_byte_limiter(
        self._restore_concurrent_bytes
    )
    param_infos = jax.tree.map(
        lambda info: dataclasses.replace(info, byte_limiter=byte_limiter),
        param_infos,
    )

    # Handle parameters from aggregate file.
    def _process_aggregated_value(meta_or_value, args):
      if not isinstance(meta_or_value, tree_metadata.ValueMetadataEntry):
        meta_or_value = _try_array_cast(meta_or_value, args.dtype)
        meta_or_value = _maybe_shard_array(meta_or_value, args)
      return meta_or_value

    flat_aggregate = tree_utils.to_flat_dict(
        jax.tree_util.tree_map(
            _process_aggregated_value, metadata, restore_args
        ),
    )

    batch_requests = (
        base_pytree_checkpoint_handler.batched_serialization_requests(
            metadata,
            param_infos,
            restore_args,
            self._type_handler_registry,
        )
    )
    deserialized_batches = []
    deserialized_batches_ops = []
    for request in batch_requests:
      deserialized_batches_ops.append(
          request.handler.deserialize(request.infos, request.args)
      )
    deserialized_batches += await asyncio.gather(*deserialized_batches_ops)

    flat_restored = {}
    for request, deserialized in zip(batch_requests, deserialized_batches):
      for key, value in zip(request.keys, deserialized):
        flat_restored[key] = value
    # Add in any values which were not deserialized, coming from aggregate file.
    for key in flat_aggregate.keys():
      if key not in flat_restored:
        flat_restored[key] = flat_aggregate[key]
    return tree_utils.from_flat_dict(flat_restored, target=item)

  def restore(
      self,
      directory: epath.Path,
      item: Optional[PyTree] = None,
      restore_args: Optional[PyTree] = None,
      transforms: Optional[PyTree] = None,
      transforms_default_to_original: bool = True,
      legacy_transform_fn: Optional[LegacyTransformFn] = None,
      args: Optional[PyTreeRestoreArgs] = None,
  ) -> PyTree:
    """Restores a PyTree from the checkpoint directory at the given path.

    In the most basic case, only `directory` is required. The tree will be
    restored exactly as saved, and all leaves will be restored as the correct
    types (assuming the tree metadata is present).

    However, `restore_args` is often required as well. This PyTree gives a
    `RestoreArgs` object (or subclass) for every leaf in the tree. Many types,
    such as string or `np.ndarray` do not require any special options for
    restoration. When restoring an individual leaf as `jax.Array`, however,
    some properties may be required.

    One example is `sharding`, which defines how a `jax.Array` in the restored
    tree should be partitioned. `mesh` and `mesh_axes` can also be used to
    specify `sharding`, but `sharding` is the preferred way of specifying this
    partition since `mesh` and `mesh_axes` only constructs
    `jax.sharding.NamedSharding`. For more information, see `ArrayTypeHandler`
    documentation and JAX sharding documentation.

    Example::

      ckptr = ocp.Checkpointer(ocp.PyTreeCheckpointHandler())
      restore_args = {
          'layer0': {
              'w': ocp.RestoreArgs(),
              'b': ocp.RestoreArgs(),
          },
          'layer1': {
              'w': ocp.ArrayRestoreArgs(
                  # Restores as jax.Array, regardless of how it was saved.
                  restore_type=jax.Array,
                  sharding=jax.sharding.Sharding(...),
                  # Warning: may truncate or pad!
                  global_shape=(x, y),
                ),
              'b': ocp.ArrayRestoreArgs(
                  restore_type=jax.Array,
                  sharding=jax.sharding.Sharding(...),
                  global_shape=(x, y),
                ),
          },
      }
      ckptr.restore(path, args=ocp.PyTreeRestore(restore_args=restore_args))

    Providing `item` is typically only necessary when restoring a custom PyTree
    class (or when using transformations). In this case, the restored object
    will take on the same structure as `item`.

    Example::

      @flax.struct.dataclass
      class TrainState:
        layer0: dict[str, jax.Array]
        layer1: dict[str, jax.Array]

      ckptr = Checkpointer(PyTreeCheckpointHandler())
      train_state = TrainState(
          layer0={
              'w': jax.Array(...),  # zeros
              'b': jax.Array(...),  # zeros
          },
          layer1={
              'w': jax.Array(...),  # zeros
              'b': jax.Array(...),  # zeros
          },
      )
      restore_args = jax.tree.map(_make_restore_args, train_state)
      ckptr.restore(path, item=train_state, restore_args=restore_args)
      # restored tree is of type `TrainState`.

    Args:
      directory: saved checkpoint location directory.
      item: Deprecated, use `args`.
      restore_args: Deprecated, use `args`.
      transforms: Deprecated, use `args`.
      transforms_default_to_original: See transform_utils.apply_transformations.
      legacy_transform_fn: Deprecated, use `args`.
      args: `PyTreeRestoreArgs` (see below).

    Returns:
      A PyTree matching the structure of `item`.

    Raises:
      FileNotFoundError: `directory` does not exist or is missing required files
      ValueError: `transforms` is provided without `item`.
      ValueError: `transforms` contains elements with `multi_value_fn`.
    """
    if self._pytree_metadata_options.support_rich_types:
      raise NotImplementedError(
          'Restore is not supported for rich typed metadata yet. Please set'
          ' PyTreeMetadataOptions.support_rich_types=False.'
      )
    if not directory.exists():
      raise FileNotFoundError(
          f'Requested directory for restore does not exist at {directory}.'
      )
    if isinstance(item, CheckpointArgs):
      raise ValueError(
          'Make sure to specify kwarg name `args=` when providing'
          ' `PyTreeRestoreArgs`.'
      )
    if args is None:
      args = PyTreeRestoreArgs(
          item,
          restore_args,
          transforms,
          transforms_default_to_original,
          legacy_transform_fn,
      )
    item = args.item
    restore_args = args.restore_args
    transforms = args.transforms
    transforms_default_to_original = args.transforms_default_to_original
    legacy_transform_fn = args.legacy_transform_fn

    try:
      can_ignore_aggregate_file = utils.all_leaves_are_placeholders(
          self._read_aggregate_file(directory)
      )
    except FileNotFoundError:
      can_ignore_aggregate_file = True

    # Delegate to `BasePyTreeCheckpointHandler` as long as transformation
    # options are not specified and metadata file exists and we do not need to
    # read from aggregate file.
    if (
        (directory / PYTREE_METADATA_FILE).exists()
        and can_ignore_aggregate_file
        and transforms is None
        and legacy_transform_fn is None
    ):
      args = BasePyTreeRestoreArgs(
          item,
          restore_args=restore_args,
      )
      return self._handler_impl.restore(directory, args=args)

    logging.vlog(1, 'directory=%s, restore_args=%s', directory, restore_args)
    if not directory.exists():
      raise FileNotFoundError(
          f'Requested directory for restore does not exist at {directory}'
      )
    structure, use_zarr3_metadata = self._get_internal_metadata(directory)
    # `checkpoint_restore_args` has a structure relative to the checkpoint,
    # while `restore_args` remains structured relative to the output.

    use_zarr3 = (
        use_zarr3_metadata
        if use_zarr3_metadata is not None
        else self._use_zarr3
    )
    param_infos, checkpoint_restore_args = _get_restore_parameters(
        directory,
        item,
        structure,
        self._handler_impl.get_param_names(structure),
        transforms,
        restore_args,
        self._pytree_metadata_options,
        transforms_default_to_original=transforms_default_to_original,
        use_zarr3=use_zarr3,
    )

    if legacy_transform_fn is not None and transforms is not None:
      raise ValueError(
          'Cannot provide both `transforms` and `legacy_transform_fn`.'
      )
    if legacy_transform_fn is not None:
      structure, param_infos = legacy_transform_fn(item, structure, param_infos)
      if restore_args is None:
        restore_args = jax.tree.map(lambda x: RestoreArgs(), item)
      checkpoint_restore_args = restore_args

    def _maybe_set_default_restore_types(value_meta: Any, arg: RestoreArgs):
      if (
          isinstance(value_meta, tree_metadata.ValueMetadataEntry)
          and not value_meta.skip_deserialize
          and value_meta.value_type == empty_values.RESTORE_TYPE_UNKNOWN
      ):
        return dataclasses.replace(
            value_meta, value_type=type_handlers.default_restore_type(arg)
        )
      return value_meta

    # If metadata file was missing in the checkpoint, we need to decide
    # restore_type based on RestoreArgs.
    structure = jax.tree.map(
        _maybe_set_default_restore_types, structure, checkpoint_restore_args
    )

    restored_item = asyncio_utils.run_sync(
        self._maybe_deserialize(
            structure, structure, param_infos, checkpoint_restore_args
        )
    )

    if not legacy_transform_fn:
      restored_item = _transform_checkpoint(
          item,
          restored_item,
          restore_args,
          transforms,
          transforms_default_to_original,
      )

    if logging.vlog_is_on(1):
      logging.vlog(1, 'param_infos: %s', param_infos)
      logging.vlog(1, 'checkpoint_restore_args: %s', checkpoint_restore_args)
      logging.vlog(1, 'restored_item: %s', jax.tree.structure(restored_item))
      logging.vlog(
          1,
          'ts_metrics: %s',
          json.dumps(ts.experimental_collect_matching_metrics('/tensorstore/')),
      )

    return restored_item

  def _read_aggregate_file(self, directory: epath.Path) -> PyTree:
    """Restores the aggregate file representing PyTree structure."""
    checkpoint_path = directory / self._aggregate_filename
    if checkpoint_path.exists():
      return self._aggregate_handler.deserialize(checkpoint_path)
    elif self._use_ocdbt:
      raise FileNotFoundError(
          f'Checkpoint structure file does not exist at {directory}.'
      )
    else:
      return utils.pytree_structure(directory)

  def _get_internal_metadata(
      self, directory: epath.Path
  ) -> Tuple[PyTree, Optional[bool]]:
    """Gets limited information needed to fully restore the checkpoint.

    This information just consists of the restore type for each leaf, as well
    as the aggregated value (from the msgpack file) if present, and determines
    whether we need to deserialize the parameter using TypeHandler later.

    Args:
      directory: directory

    Returns:
      A PyTree with leaves of ValueMetadataEntry or real values if restored from
      the aggregate file (or if empty nodes).

    Raises:
      FileNotFoundError: no structure could be identified for the checkpoint at
      `directory`.
    """
    # Try reading metadata file.
    try:
      internal_tree_metadata = self._handler_impl._read_metadata_file(directory)  # pylint: disable=protected-access
      use_zarr3 = internal_tree_metadata.use_zarr3
      value_metadata_tree = internal_tree_metadata.as_nested_tree()
      flat_value_metadatas = tree_utils.to_flat_dict(
          value_metadata_tree, keep_empty_nodes=True
      )
    except FileNotFoundError:
      jax.monitoring.record_event('/jax/orbax/deprecation/missing_metadata')
      value_metadata_tree = None
      flat_value_metadatas = None
      use_zarr3 = None
    # Try reading aggregate file.
    try:
      aggregate_tree = self._read_aggregate_file(directory)
      flat_aggregate = tree_utils.to_flat_dict(
          aggregate_tree, keep_empty_nodes=True
      )
    except FileNotFoundError:
      jax.monitoring.record_event('/jax/orbax/deprecation/missing_structure')
      aggregate_tree = None
      flat_aggregate = None

    def _is_empty_value(value):
      return empty_values.is_supported_empty_value(
          value, self._pytree_metadata_options
      ) or not utils.leaf_is_placeholder(value)

    def _process_aggregate_leaf(value):
      if _is_empty_value(value):
        return value
      return tree_metadata.ValueMetadataEntry(
          value_type=empty_values.RESTORE_TYPE_UNKNOWN,
          skip_deserialize=False,
      )

    def _process_metadata_and_aggregate_leaves(value_meta, value):
      if _is_empty_value(value):
        return value
      if empty_values.is_empty_typestr(value_meta.value_type):
        return empty_values.get_empty_value_from_typestr(
            value_meta.value_type, self._pytree_metadata_options
        )
      return value_meta

    # Handle cases of missing metadata and/or aggregate files.
    structure_tree = value_metadata_tree or aggregate_tree
    if flat_value_metadatas is None and flat_aggregate is None:
      raise FileNotFoundError(
          f'No structure could be identified for the checkpoint at {directory}.'
      )
    elif flat_value_metadatas is None:
      # Metadata file is missing. This is an older checkpoint.
      # TODO(b/353310784) Track usages.
      flat_structure = jax.tree_util.tree_map(
          _process_aggregate_leaf,
          flat_aggregate,
          is_leaf=tree_utils.is_empty_or_leaf,
      )
    elif flat_aggregate is None:
      # Aggregate file is missing, so we can just use the metadata_tree as the
      # structure. This is a newer checkpoint.
      return value_metadata_tree, use_zarr3
    else:
      # Avoid tree_map because input trees may be mismatched (due to empty
      # values missing from msgpack structure).
      flat_structure = {}
      for tuple_key in flat_value_metadatas.keys():
        value_meta = flat_value_metadatas[tuple_key]
        if tuple_key in flat_aggregate:
          flat_structure[tuple_key] = _process_metadata_and_aggregate_leaves(
              value_meta, flat_aggregate[tuple_key]
          )
        else:
          if empty_values.is_empty_typestr(value_meta.value_type):
            flat_structure[tuple_key] = (
                empty_values.get_empty_value_from_typestr(
                    value_meta.value_type, self._pytree_metadata_options
                )
            )
          else:
            flat_structure[tuple_key] = value_meta

    return (
        tree_utils.from_flat_dict(flat_structure, target=structure_tree),
        use_zarr3,
    )

  def metadata(self, directory: epath.Path) -> tree_metadata.TreeMetadata:
    """Returns tree metadata.

    The result will be a PyTree matching the structure of the saved checkpoint.
    Note that if the item saved was a custom class, the restored metadata will
    be returned as a nested dictionary representation.

    Example::

      {
        'layer0': {
            'w': ArrayMetadata(dtype=jnp.float32, shape=(8, 8), shards=(1, 2)),
            'b': ArrayMetadata(dtype=jnp.float32, shape=(8,), shards=(1,)),
        },
        'step': ScalarMetadata(dtype=jnp.int64),
      }

    If the required metadata file is not present, this method will raise an
    error.

    Args:
      directory: checkpoint location.

    Returns:
      tree containing metadata.
    """
    return self._handler_impl.metadata(directory)

  def finalize(self, directory: epath.Path) -> None:
    """Finalization step.

    Called automatically by the Checkpointer/AsyncCheckpointer just before the
    checkpoint is considered "finalized" in the sense of ensuring atomicity. See
    documentation for `type_handlers.merge_ocdbt_per_process_files`.

    Args:
      directory: Path where the checkpoint is located.
    """
    self._handler_impl.finalize(directory)

  def close(self):
    """Closes the handler. Called automatically by Checkpointer."""
    self._handler_impl.close()


@register_with_handler(PyTreeCheckpointHandler, for_save=True)
@dataclasses.dataclass
class PyTreeSaveArgs(CheckpointArgs):
  """Parameters for saving a PyTree.

  Attributes:
    item (required): a PyTree to be saved.
    save_args: a PyTree with the same structure of `item`, which consists of
      `ocp.SaveArgs` objects as values. `None` can be used for values where no
      `SaveArgs` are specified.
    ocdbt_target_data_file_size: Specifies the target size (in bytes) of each
      OCDBT data file.  It only applies when OCDBT is enabled and Zarr3 must be
      turned on.  If left unspecified, default size is 2GB.  A value of 0
      indicates no maximum file size limit.  For best results, ensure
      chunk_byte_size is smaller than this value.  For more details, refer to
      https://google.github.io/tensorstore/kvstore/ocdbt/index.html#json-kvstore/ocdbt.target_data_file_size
    enable_pinned_host_transfer: If False, disables transfer to pinned host when
      copying from device to host, regardless of the presence of pinned host
      memory.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """

  item: PyTree
  save_args: Optional[PyTree] = None
  ocdbt_target_data_file_size: Optional[int] = None
  enable_pinned_host_transfer: bool = False
  custom_metadata: tree_types.JsonType | None = None

  def __post_init__(self):
    if isinstance(self.item, tree_metadata.TreeMetadata):
      raise ValueError('Cannot save TreeMetadata.')


@register_with_handler(PyTreeCheckpointHandler, for_restore=True)
@dataclasses.dataclass
class PyTreeRestoreArgs(CheckpointArgs):
  """Parameters for restoring a PyTree.

  Attributes (all optional):
    item: provides the tree structure for the restored item. If not provided,
      will infer the structure from the saved checkpoint. Transformations will
      not be run in this case. Necessary particularly in the case where the
      caller needs to restore the tree as a custom object.
      `TreeMetadata` is also allowed as the tree used to
      define the restored structure.
    restore_args: optional object containing additional arguments for
      restoration. It should be a PyTree matching the structure of `item`, or
      if `item` is not provided, then it should match the structure of the
      checkpoint. Each value in the tree should be a `RestoreArgs` object (OR
      a subclass of `RestoreArgs`). Importantly, note that when restoring a
      leaf as a certain type, a specific subclass of `RestoreArgs` may be
      required. `RestoreArgs` also provides the option to customize the
      restore type of an individual leaf.
      `TreeMetadata` is also allowed as the `restore_args` tree.
    transforms: a PyTree of transformations that should be applied to the
      saved tree in order to obtain a final structure. The `transforms` tree
      structure should conceptually match that of `item`, but the use of
      regexes and implicit keys means that it does not need to match
      completely. See `transform_utils` for further information.
      `TreeMetadata` is also allowed as the `transforms` tree.
    transforms_default_to_original: See transform_utils.apply_transformations.
    legacy_transform_fn: WARNING: NOT GENERALLY SUPPORTED. A function which
      accepts the `item` argument, a PyTree checkpoint structure and a PyTree
      of ParamInfos based on the checkpoint. Returns a transformed PyTree
      matching the desired return tree structure, and a matching ParamInfo
      tree.
  """

  item: Optional[PyTree] = None
  restore_args: Optional[PyTree] = None
  transforms: Optional[PyTree] = None
  transforms_default_to_original: bool = True
  legacy_transform_fn: Optional[LegacyTransformFn] = None

  def __post_init__(self):
    if isinstance(self.item, tree_metadata.TreeMetadata):
      self.item = self.item.tree
    if isinstance(self.restore_args, tree_metadata.TreeMetadata):
      self.restore_args = self.restore_args.tree
    if isinstance(self.transforms, tree_metadata.TreeMetadata):
      self.transforms = self.transforms.tree



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Testing utilities for pytrees."""

from __future__ import annotations

import dataclasses
from typing import Any, Callable, Dict, List, Mapping, NamedTuple, Sequence, Tuple

import chex
import flax
import jax
import jax.numpy as jnp
import numpy as np
import optax
from orbax.checkpoint._src.metadata import tree as tree_metadata
from orbax.checkpoint._src.metadata import tree_rich_types

PyTree = Any


def create_namedtuple(
    cls,
    field_value_tuples: list[tuple[str, tree_metadata.ValueMetadataEntry]],
) -> type[tuple[Any, ...]]:
  """Returns instance of a new namedtuple type structurally identical to `cls`."""
  fields, values = zip(*field_value_tuples)
  module_name, class_name = tree_rich_types._module_and_class_name(cls)  # pylint: disable=protected-access
  new_type = tree_rich_types._new_namedtuple_type(module_name, class_name, fields)  # pylint: disable=protected-access
  return new_type(*values)


class MuNu(NamedTuple):
  mu: jax.Array | None
  nu: np.ndarray | None


class IntegerNamedTuple(NamedTuple):
  x: int | None
  y: int | None


class EmptyNamedTuple(NamedTuple):
  pass


class NamedTupleWithNestedAttributes(NamedTuple):
  nested_mu_nu: MuNu | None = None
  nested_dict: Dict[str, jax.Array] | None = None
  nested_tuple: Tuple[jax.Array, jax.Array] | None = None
  nested_empty_named_tuple: EmptyNamedTuple | None = None
  my_empty_chex: MyEmptyChex | None = None


@chex.dataclass
class MyEmptyChex:
  pass


@chex.dataclass
class MyChex:
  my_jax_array: jax.Array | None = None
  my_np_array: np.ndarray | None = None


@chex.dataclass
class MyChexWithNestedAttributes:
  my_chex: MyChex | None = None
  my_dict: Dict[str, jax.Array] | None = None
  my_list: List[np.ndarray] | None = None
  my_empty_chex: MyEmptyChex | None = None


@flax.struct.dataclass
class MyEmptyFlax(flax.struct.PyTreeNode):
  pass


@flax.struct.dataclass
class MyFlax(flax.struct.PyTreeNode):
  my_jax_array: jax.Array | None = None
  my_nested_mapping: Mapping[str, Any] | None = None
  my_sequence: Sequence[Any] | None = None


@dataclasses.dataclass
class MyEmptyDataClass:
  pass


@dataclasses.dataclass
class MyDataClass:
  my_jax_array: jax.Array | None = None
  my_np_array: np.ndarray | None = None
  my_empty_dataclass: MyEmptyDataClass | None = None
  my_chex: MyChex | None = None


class MyEmptyClass:
  pass


class MyClass:
  """Test class.

  Attributes:
    a: optional jax.Array. default=None.
    b: optional np.ndarray. default=None.
  """

  def __init__(
      self,
      a: jax.Array | None = None,
      b: np.ndarray | None = None,
  ):
    self._a = a
    self._b = b

  @property
  def a(self) -> jax.Array | None:
    return self._a

  @property
  def b(self) -> np.ndarray | None:
    return self._b


class MyClassWithNestedAttributes:
  """Test class.

  Attributes:
    my_empty_class: optional `MyEmptyClass`. default=None.
    my_class: optional `MyClass`. default=None.
    my_chex: optional `MyChex`. default=None.
    mu_nu: optional `MuNu`. default=None.
  """

  def __init__(
      self,
      my_empty_class: MyEmptyClass | None = None,
      my_class: MyClass | None = None,
      my_chex: MyChex | None = None,
      mu_nu: MuNu | None = None,
  ):
    self._my_empty_class = my_empty_class
    self._my_class = my_class
    self._my_chex = my_chex
    self._mu_nu = mu_nu


@dataclasses.dataclass
class TestPyTree:
  """Test data class for pytree.

  Attributes:
    unique_name: unique name for the test.
    provide_tree: function to provide the pytree.
    expected_save_response: expected save response. Can be a BaseException or
      None. None implies that save should succeed.
    expected_restore_response: expected restore response. Can be a
      BaseException, a function to provide the pytree or None. None implies that
      expected restored tree should be the same as the tree returned by
      `provide_tree` function.
    expected_nested_tree_metadata: PyTree of ValueMetadataEntry as returned by
      `InternalTreeMetadata.as_nested_tree` with
      `PyTreeMetadataOptions.support_rich_types=false`.
    expected_nested_tree_metadata_with_rich_types: PyTree of ValueMetadataEntry
      as returned by `InternalTreeMetadata.as_nested_tree` with
      `PyTreeMetadataOptions.support_rich_types=true`. If not provided,
      `expected_nested_tree_metadata` will be used.
  """

  unique_name: str
  # Provide tree lazily via `tree_provider` to avoid error:
  # RuntimeError: Attempted call to JAX before absl.app.run() is called.
  provide_tree: Callable[[], PyTree]
  expected_save_response: BaseException | None = None
  expected_restore_response: BaseException | Callable[[], PyTree] | None = None
  expected_nested_tree_metadata: PyTree | None = None
  expected_nested_tree_metadata_with_rich_types: PyTree | None = None

  def __post_init__(self):
    self.expected_restore_response = (
        self.expected_restore_response or self.provide_tree
    )
    self.expected_nested_tree_metadata_with_rich_types = (
        self.expected_nested_tree_metadata_with_rich_types
        or self.expected_nested_tree_metadata
    )

  def __str__(self):
    return self.unique_name

  def __repr__(self):
    return self.unique_name


TEST_PYTREES = [
    TestPyTree(
        unique_name='empty_pytree',
        provide_tree=lambda: {},
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata=(
            tree_metadata.ValueMetadataEntry(
                value_type='Dict',
                skip_deserialize=True,
            )
        ),
    ),
    TestPyTree(
        unique_name='empty_dict',
        provide_tree=lambda: {'empty_dict': {}},
        expected_nested_tree_metadata={
            'empty_dict': tree_metadata.ValueMetadataEntry(
                value_type='Dict',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='empty_list',
        provide_tree=lambda: {'empty_list': []},
        expected_nested_tree_metadata={
            'empty_list': tree_metadata.ValueMetadataEntry(
                value_type='List',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='empty_tuple',
        provide_tree=lambda: {'empty_tuple': tuple()},
        expected_nested_tree_metadata={
            'empty_tuple': tree_metadata.ValueMetadataEntry(
                value_type='Tuple',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='empty_named_tuple',
        provide_tree=lambda: {'empty_named_tuple': EmptyNamedTuple()},
        expected_nested_tree_metadata={
            'empty_named_tuple': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
        expected_nested_tree_metadata_with_rich_types={
            'empty_named_tuple': tree_metadata.ValueMetadataEntry(
                value_type='NamedTuple',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='tuple_of_empty_list',
        provide_tree=lambda: {'tuple_of_empty_list': ([],)},
        expected_nested_tree_metadata={
            'tuple_of_empty_list': [
                tree_metadata.ValueMetadataEntry(
                    value_type='List',
                    skip_deserialize=True,
                ),
            ]
        },
        expected_nested_tree_metadata_with_rich_types={
            'tuple_of_empty_list': tuple([
                tree_metadata.ValueMetadataEntry(
                    value_type='List',
                    skip_deserialize=True,
                ),
            ])
        },
    ),
    TestPyTree(
        unique_name='list_of_empty_tuple',
        provide_tree=lambda: {'list_of_empty_tuple': [tuple()]},
        expected_nested_tree_metadata={
            'list_of_empty_tuple': [
                tree_metadata.ValueMetadataEntry(
                    value_type='Tuple',
                    skip_deserialize=True,
                ),
            ]
        },
    ),
    TestPyTree(
        unique_name='none_param',
        provide_tree=lambda: {'none_param': None},
        expected_nested_tree_metadata={
            'none_param': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='scalar_param',
        provide_tree=lambda: {'scalar_param': 1},
        expected_nested_tree_metadata={
            'scalar_param': tree_metadata.ValueMetadataEntry(
                value_type='scalar',
                skip_deserialize=False,
            )
        },
    ),
    TestPyTree(
        unique_name='nested_scalars',
        provide_tree=lambda: {
            'b': {'scalar_param': 1, 'nested_scalar_param': 3.4}
        },
        expected_nested_tree_metadata={
            'b': {
                'scalar_param': tree_metadata.ValueMetadataEntry(
                    value_type='scalar',
                    skip_deserialize=False,
                ),
                'nested_scalar_param': tree_metadata.ValueMetadataEntry(
                    value_type='scalar',
                    skip_deserialize=False,
                ),
            }
        },
    ),
    TestPyTree(
        unique_name='list_of_arrays',
        provide_tree=lambda: {'list_of_arrays': [np.arange(8), jnp.arange(8)]},
        expected_nested_tree_metadata={
            'list_of_arrays': [
                tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
            ]
        },
    ),
    TestPyTree(
        unique_name='dict_of_nested_data',
        provide_tree=lambda: {
            'x': {'a': np.arange(8), 'b': MyEmptyChex()},
            'y': (np.arange(8), MyChex(my_np_array=np.arange(8))),
            'z': [
                {'c': np.arange(8)},
                [
                    (np.arange(8),),
                ],
            ],
        },
        expected_nested_tree_metadata={
            'x': {
                'a': tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                'b': tree_metadata.ValueMetadataEntry(
                    value_type='Dict',
                    skip_deserialize=True,
                ),
            },
            'y': [
                tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                {
                    'my_jax_array': tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    'my_np_array': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
            ],
            'z': [
                {
                    'c': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    )
                },
                [
                    [
                        tree_metadata.ValueMetadataEntry(
                            value_type='np.ndarray',
                            skip_deserialize=False,
                        ),
                    ],
                ],
            ],
        },
        expected_nested_tree_metadata_with_rich_types={
            'x': {
                'a': tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                'b': tree_metadata.ValueMetadataEntry(
                    value_type='Dict',
                    skip_deserialize=True,
                ),
            },
            'y': (
                tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                {
                    'my_jax_array': tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    'my_np_array': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
            ),
            'z': [
                {
                    'c': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    )
                },
                [
                    (
                        tree_metadata.ValueMetadataEntry(
                            value_type='np.ndarray',
                            skip_deserialize=False,
                        ),
                    ),
                ],
            ],
        },
    ),
    TestPyTree(
        unique_name='tuple_of_arrays',
        provide_tree=lambda: {'tuple_of_arrays': (np.arange(8), jnp.arange(8))},
        expected_nested_tree_metadata={
            'tuple_of_arrays': [
                tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
            ]
        },
        expected_nested_tree_metadata_with_rich_types={
            'tuple_of_arrays': tuple([
                tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
                tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
            ])
        },
    ),
    TestPyTree(
        unique_name='mu_nu',
        provide_tree=lambda: {'mu_nu': MuNu(mu=jnp.arange(8), nu=np.arange(8))},
        expected_nested_tree_metadata={
            'mu_nu': {
                'mu': tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
                'nu': tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
            }
        },
        expected_nested_tree_metadata_with_rich_types={
            'mu_nu': create_namedtuple(
                MuNu,
                [
                    (
                        'mu',
                        tree_metadata.ValueMetadataEntry(
                            value_type='jax.Array',
                            skip_deserialize=False,
                        ),
                    ),
                    (
                        'nu',
                        tree_metadata.ValueMetadataEntry(
                            value_type='np.ndarray',
                            skip_deserialize=False,
                        ),
                    ),
                ],
            )
        },
    ),
    TestPyTree(
        unique_name='my_empty_chex',
        provide_tree=lambda: {'my_empty_chex': MyEmptyChex()},
        expected_nested_tree_metadata={
            'my_empty_chex': tree_metadata.ValueMetadataEntry(
                value_type='Dict',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_my_chex',
        provide_tree=lambda: {'default_my_chex': MyChex()},
        expected_nested_tree_metadata={
            'default_my_chex': {
                'my_jax_array': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_np_array': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
            }
        },
    ),
    TestPyTree(
        unique_name='my_chex',
        provide_tree=lambda: {
            'my_chex': MyChex(
                my_jax_array=jnp.arange(8),
                my_np_array=np.arange(8),
            )
        },
        expected_nested_tree_metadata={
            'my_chex': {
                'my_jax_array': tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
                'my_np_array': tree_metadata.ValueMetadataEntry(
                    value_type='np.ndarray',
                    skip_deserialize=False,
                ),
            }
        },
    ),
    TestPyTree(
        unique_name='default_my_chex_with_nested_attrs',
        provide_tree=lambda: {
            'default_my_chex_with_nested_attrs': MyChexWithNestedAttributes()
        },
        expected_nested_tree_metadata={
            'default_my_chex_with_nested_attrs': {
                'my_chex': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_dict': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_empty_chex': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_list': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
            }
        },
    ),
    TestPyTree(
        unique_name='my_chex_with_nested_attrs',
        provide_tree=lambda: {
            'my_chex_with_nested_attrs': MyChexWithNestedAttributes(
                my_chex=MyChex(
                    my_jax_array=jnp.arange(8), my_np_array=np.arange(8)
                ),
                my_dict={'a': jnp.arange(8), 'b': np.arange(8)},
                my_list=[jnp.arange(8), np.arange(8)],
                my_empty_chex=MyEmptyChex(),
            )
        },
        expected_nested_tree_metadata={
            'my_chex_with_nested_attrs': {
                'my_chex': {
                    'my_jax_array': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'my_np_array': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
                'my_dict': {
                    'a': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'b': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
                'my_empty_chex': tree_metadata.ValueMetadataEntry(
                    value_type='Dict',
                    skip_deserialize=True,
                ),
                'my_list': [
                    tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                ],
            }
        },
    ),
    TestPyTree(
        unique_name='my_empty_dataclass',
        provide_tree=lambda: {'my_empty_dataclass': MyEmptyDataClass()},
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'my_empty_dataclass': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_my_dataclass',
        provide_tree=lambda: {'default_my_dataclass': MyDataClass()},
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'default_my_dataclass': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='my_dataclass',
        provide_tree=lambda: {
            'my_dataclass': MyDataClass(
                my_jax_array=jnp.arange(8),
                my_np_array=np.arange(8),
                my_empty_dataclass=MyEmptyDataClass(),
                my_chex=MyChex(
                    my_jax_array=jnp.arange(8),
                    my_np_array=np.arange(8),
                ),
            )
        },
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'my_dataclass': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='my_empty_class',
        provide_tree=lambda: {'my_empty_class': MyEmptyClass()},
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'my_empty_class': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_my_class',
        provide_tree=lambda: {'default_my_class': MyClass()},
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'default_my_class': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='my_class',
        provide_tree=lambda: {
            'my_class': MyClass(a=jnp.arange(8), b=np.arange(8))
        },
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'my_class': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_my_class_with_nested_attrs',
        provide_tree=lambda: {
            'default_my_class_with_nested_attrs': MyClassWithNestedAttributes()
        },
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'default_my_class_with_nested_attrs': (
                tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                )
            )
        },
    ),
    TestPyTree(
        unique_name='my_class_with_nested_attrs',
        provide_tree=lambda: {
            'my_class_with_nested_attrs': MyClassWithNestedAttributes(
                my_empty_class=MyEmptyClass(),
                my_class=MyClass(a=jnp.arange(8), b=np.arange(8)),
                my_chex=MyChex(
                    my_jax_array=jnp.arange(8), my_np_array=np.arange(8)
                ),
                mu_nu=MuNu(mu=jnp.arange(8), nu=np.arange(8)),
            )
        },
        expected_save_response=ValueError(),
        expected_restore_response=ValueError(),
        expected_nested_tree_metadata={
            'my_class_with_nested_attrs': tree_metadata.ValueMetadataEntry(
                value_type='None',
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_named_tuple_with_nested_attrs',
        provide_tree=lambda: {
            'default_named_tuple_with_nested_attrs': (
                NamedTupleWithNestedAttributes()
            )
        },
        expected_nested_tree_metadata={
            'default_named_tuple_with_nested_attrs': {
                'nested_mu_nu': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'nested_dict': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'nested_tuple': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'nested_empty_named_tuple': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_empty_chex': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
            }
        },
        expected_nested_tree_metadata_with_rich_types={
            'default_named_tuple_with_nested_attrs': create_namedtuple(
                NamedTupleWithNestedAttributes,
                [
                    (
                        'nested_mu_nu',
                        tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    ),
                    (
                        'nested_dict',
                        tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    ),
                    (
                        'nested_tuple',
                        tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    ),
                    (
                        'nested_empty_named_tuple',
                        tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    ),
                    (
                        'my_empty_chex',
                        tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    ),
                ],
            )
        },
    ),
    TestPyTree(
        unique_name='named_tuple_with_nested_attrs',
        provide_tree=lambda: {
            'named_tuple_with_nested_attrs': NamedTupleWithNestedAttributes(
                nested_mu_nu=MuNu(mu=jnp.arange(8), nu=np.arange(8)),
                nested_dict={'a': jnp.arange(8), 'b': np.arange(8)},
                nested_tuple=(jnp.arange(8), jnp.arange(8)),
                nested_empty_named_tuple=EmptyNamedTuple(),
                my_empty_chex=MyEmptyChex(),
            )
        },
        expected_nested_tree_metadata={
            'named_tuple_with_nested_attrs': {
                'nested_mu_nu': {
                    'mu': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'nu': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
                'nested_dict': {
                    'a': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'b': tree_metadata.ValueMetadataEntry(
                        value_type='np.ndarray',
                        skip_deserialize=False,
                    ),
                },
                'nested_tuple': [
                    tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                ],
                'nested_empty_named_tuple': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_empty_chex': tree_metadata.ValueMetadataEntry(
                    value_type='Dict',
                    skip_deserialize=True,
                ),
            }
        },
        expected_nested_tree_metadata_with_rich_types={
            'named_tuple_with_nested_attrs': create_namedtuple(
                NamedTupleWithNestedAttributes,
                [
                    (
                        'nested_mu_nu',
                        create_namedtuple(
                            MuNu,
                            [
                                (
                                    'mu',
                                    tree_metadata.ValueMetadataEntry(
                                        value_type='jax.Array',
                                        skip_deserialize=False,
                                    ),
                                ),
                                (
                                    'nu',
                                    tree_metadata.ValueMetadataEntry(
                                        value_type='np.ndarray',
                                        skip_deserialize=False,
                                    ),
                                ),
                            ],
                        ),
                    ),
                    (
                        'nested_dict',
                        {
                            'a': tree_metadata.ValueMetadataEntry(
                                value_type='jax.Array',
                                skip_deserialize=False,
                            ),
                            'b': tree_metadata.ValueMetadataEntry(
                                value_type='np.ndarray',
                                skip_deserialize=False,
                            ),
                        },
                    ),
                    (
                        'nested_tuple',
                        (
                            tree_metadata.ValueMetadataEntry(
                                value_type='jax.Array',
                                skip_deserialize=False,
                            ),
                            tree_metadata.ValueMetadataEntry(
                                value_type='jax.Array',
                                skip_deserialize=False,
                            ),
                        ),
                    ),
                    (
                        'nested_empty_named_tuple',
                        tree_metadata.ValueMetadataEntry(
                            value_type='NamedTuple',
                            skip_deserialize=True,
                        ),
                    ),
                    (
                        'my_empty_chex',
                        tree_metadata.ValueMetadataEntry(
                            value_type='Dict',
                            skip_deserialize=True,
                        ),
                    ),
                ],
            )
        },
    ),
    TestPyTree(
        unique_name='my_empty_flax',
        provide_tree=lambda: {'my_empty_flax': MyEmptyFlax()},
        expected_nested_tree_metadata={
            'my_empty_flax': tree_metadata.ValueMetadataEntry(
                value_type='None',  # TODO: b/378905913 - Should be Dict.
                skip_deserialize=True,
            )
        },
    ),
    TestPyTree(
        unique_name='default_my_flax',
        provide_tree=lambda: {'default_my_flax': MyFlax()},
        expected_nested_tree_metadata={
            'default_my_flax': {
                'my_jax_array': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_nested_mapping': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
                'my_sequence': tree_metadata.ValueMetadataEntry(
                    value_type='None',
                    skip_deserialize=True,
                ),
            }
        },
    ),
    TestPyTree(
        unique_name='my_flax',
        provide_tree=lambda: {
            'my_flax': MyFlax(
                my_jax_array=jnp.arange(8),
                my_nested_mapping={
                    'a': None,
                    'b': jnp.arange(8),
                    'c': MyFlax(),
                    'd': MyFlax(
                        my_nested_mapping={
                            'e': MyEmptyChex(),
                        },
                        my_sequence=[
                            MyChex(my_jax_array=jnp.arange(8)),
                        ],
                    ),
                },
                my_sequence=(
                    None,
                    optax.EmptyState(),
                    MyEmptyFlax(),
                    MyEmptyChex(),
                    EmptyNamedTuple(),
                ),
            )
        },
        expected_nested_tree_metadata={
            'my_flax': {
                'my_jax_array': tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
                'my_nested_mapping': {
                    'a': tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    'b': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'c': {
                        'my_jax_array': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_nested_mapping': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_sequence': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    },
                    'd': {
                        'my_jax_array': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_nested_mapping': {
                            'e': tree_metadata.ValueMetadataEntry(
                                value_type='Dict',
                                skip_deserialize=True,
                            ),
                        },
                        'my_sequence': [
                            {
                                'my_jax_array': (
                                    tree_metadata.ValueMetadataEntry(
                                        value_type='jax.Array',
                                        skip_deserialize=False,
                                    )
                                ),
                                'my_np_array': tree_metadata.ValueMetadataEntry(
                                    value_type='None',
                                    skip_deserialize=True,
                                ),
                            },
                        ],
                    },
                },
                'my_sequence': [
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='Dict',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                ],
            }
        },
        expected_nested_tree_metadata_with_rich_types={
            'my_flax': {
                'my_jax_array': tree_metadata.ValueMetadataEntry(
                    value_type='jax.Array',
                    skip_deserialize=False,
                ),
                'my_nested_mapping': {
                    'a': tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    'b': tree_metadata.ValueMetadataEntry(
                        value_type='jax.Array',
                        skip_deserialize=False,
                    ),
                    'c': {
                        'my_jax_array': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_nested_mapping': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_sequence': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                    },
                    'd': {
                        'my_jax_array': tree_metadata.ValueMetadataEntry(
                            value_type='None',
                            skip_deserialize=True,
                        ),
                        'my_nested_mapping': {
                            'e': tree_metadata.ValueMetadataEntry(
                                value_type='Dict',
                                skip_deserialize=True,
                            ),
                        },
                        'my_sequence': [
                            {
                                'my_jax_array': (
                                    tree_metadata.ValueMetadataEntry(
                                        value_type='jax.Array',
                                        skip_deserialize=False,
                                    )
                                ),
                                'my_np_array': tree_metadata.ValueMetadataEntry(
                                    value_type='None',
                                    skip_deserialize=True,
                                ),
                            },
                        ],
                    },
                },
                'my_sequence': (
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='NamedTuple',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='None',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='Dict',
                        skip_deserialize=True,
                    ),
                    tree_metadata.ValueMetadataEntry(
                        value_type='NamedTuple',
                        skip_deserialize=True,
                    ),
                ),
            }
        },
    ),
]

# Suitable for parameterized.named_parameters.
TEST_PYTREES_FOR_NAMED_PARAMETERS = [
    (test_pytree.unique_name, test_pytree) for test_pytree in TEST_PYTREES
]



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at root level."""

from absl import logging
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
RootMetadata = checkpoint.RootMetadata


def serialize(metadata: RootMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""
  return {
      'custom_metadata': metadata.custom_metadata,
  }


def deserialize(metadata_dict: SerializedMetadata) -> RootMetadata:
  """Deserializes `metadata_dict` to `RootMetadata`."""
  validated_metadata_dict = {
      'custom_metadata': utils.validate_and_process_custom_metadata(
          metadata_dict.get('custom_metadata', None)
      ),
  }

  for k in metadata_dict:
    if k not in validated_metadata_dict:
      if (
          'custom_metadata' in metadata_dict
          and metadata_dict['custom_metadata']
      ):
        raise ValueError(
            'Provided metadata contains unknown key %s, and the custom_metadata'
            ' field is already defined.' % k
        )
      logging.warning(
          'Provided metadata contains unknown key %s. Adding it to'
          ' custom_metadata.',
          k,
      )
      validated_metadata_dict['custom_metadata'][k] = metadata_dict[k]

  return RootMetadata(**validated_metadata_dict)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for ensuring mesh consistency in emergency checkpoint restoration."""

import json
import time
from typing import Any, List
from absl import logging
from etils import epath
import jax
from orbax.checkpoint import options as options_lib
from orbax.checkpoint.experimental.emergency import multihost as emergency_multihost
from orbax.checkpoint.path import atomicity_defaults
from orbax.checkpoint.path import step as step_lib



_PROCESS_METADATA_FOLDER = 'process_metadata'
_PROCESS_METADATA_FILE_NAME = 'process_metadata.json'
_GLOBAL_PROCESS_METADATA_FILE_NAME = 'global_process_metadata.json'
_MESH_METADATA_FILE_NAME = 'mesh_metadata.json'

_PRIMARY_REPLICA_ID = 0
_SECONDARY_REPLICA_ID = 1

PyTree = Any


def process_metadata_folder(
    directory: epath.Path
) -> epath.Path:
  return directory


def read_process_metadata(directory: epath.Path):
  """Read process metadata from the given path."""
  metadata_folder = process_metadata_folder(directory)
  if not metadata_folder.exists():
    raise FileNotFoundError(
        f'Process metadata folder does not exist at {metadata_folder}. The'
        ' local checkpoint cannot be restored.'
    )
  if step_lib.is_tmp_checkpoint(metadata_folder):
    raise ValueError(
        f'Process metadata folder was not finalized at {metadata_folder}.'
        ' The local checkpoint cannot be restored.'
    )
  logging.info('Loading process index metadata from %s', metadata_folder)

  distributed_to_device_ids = json.loads(
      (metadata_folder / _GLOBAL_PROCESS_METADATA_FILE_NAME).read_text()
  )
  device_ids = json.loads(
      (metadata_folder / _MESH_METADATA_FILE_NAME).read_text()
  )
  return distributed_to_device_ids, device_ids


async def save_process_metadata(
    directory: epath.Path,
    global_mesh: jax.sharding.Mesh,
    distributed_to_device_ids: List[List[int]],
):
  """Saves process metadata to local storage. Runs on every process."""
  metadata_folder = process_metadata_folder(directory)
  logging.info('Saving process index metadata at %s', metadata_folder)
  if metadata_folder.exists():
    logging.warning(
        'Process metadata folder already exists at %s. Overwriting.',
        metadata_folder,
    )
    metadata_folder.rmtree()

  multiprocessing_options = options_lib.MultiprocessingOptions(
      primary_host=None
  )
  tmp_path = atomicity_defaults.get_default_temporary_path_class(
      metadata_folder
  ).from_final(metadata_folder, multiprocessing_options=multiprocessing_options)
  await tmp_path.create()

  (tmp_path.get() / _GLOBAL_PROCESS_METADATA_FILE_NAME).write_text(
      json.dumps(distributed_to_device_ids)
  )
  (tmp_path.get() / _MESH_METADATA_FILE_NAME).write_text(
      json.dumps([int(id) for id in global_mesh.device_ids.flatten()])
  )
  tmp_path.finalize(
  )


def consistent_restore_mesh_from_metadata(
    global_mesh: jax.sharding.Mesh,
    current_distributed_to_device_ids: List[List[int]],
    previous_distributed_to_device_ids: List[List[int]],
    previous_device_ids: List[int],
) -> jax.sharding.Mesh:
  """Create a mesh consistent with the saved metadata.

  Each device has its own specific local shard.
  Naive example:
    Before restart: device X (with id 0) saves shard 0.
    After restart:  device X (with new id 1) reads shard 0.
  To ensure the same device X (with different software ids) reads the
  same locally available shard and represent it in the correct index
  within the global jax.Array, we use the `restore_mesh`.

  More context on `restore_mesh`:
  1. We can think of mesh as being backed by a list of devices.
  2. Default mesh follows the default device id order [0, ..., n-1]. Or
      the user may permute it according to their needs.
  3. After restart, the user will construct the same software mesh as (2).
  4. But a given hardware device may change its id because of scheduler
      or runtime quirks.
  5. Goal: construct the mesh with the same hardware device order as
      before restart, that may not follow the current software ids.
  5. Thus, we shuffle the device order within the mesh by checking how
      each device's software ids changed across restarts.

  Args:
    global_mesh: The global mesh, provided by the user.
    current_distributed_to_device_ids: The distributed id to range of device ids
      mapping of the current incarnation.
    previous_distributed_to_device_ids: The distributed id to range of device
      ids mapping of the previous incarnation.
    previous_device_ids: The device ids of the previous incarnation.

  Returns:
    A mesh that is the same as the mesh used to save the local checkpoint.
  """
  assert isinstance(previous_device_ids, list)
  logging.info(
      'From process metadata, distributed_to_device_ids=%s',
      previous_distributed_to_device_ids,
  )
  logging.info('From process metadata, device_ids=%s', previous_device_ids)
  consistent_mesh = emergency_multihost.consistent_restore_mesh(
      jax.devices(),
      global_mesh,
      previous_device_ids,
      previous_distributed_to_device_ids=previous_distributed_to_device_ids,
      current_distributed_to_device_ids=current_distributed_to_device_ids,
  )
  logging.info(
      'Created consistent mesh with device_ids=%s',
      consistent_mesh.device_ids.flatten(),
  )
  return consistent_mesh


def consistent_restore_mesh_to_global_mesh(
    state: PyTree,
    shardings: PyTree,
) -> PyTree:
  """Transfers from consistent restore mesh to global mesh."""
  logging.info('Transferring from consistent restore mesh to global mesh')

  start_transfer = time.time()
  resharded_state = jax.device_put(state, shardings, donate=True)
  transfer_elapsed_s = time.time() - start_transfer
  logging.info(
      'Finished transferring from consistent restore mesh to global mesh'
      ' in %.2fs',
      transfer_elapsed_s,
  )
  jax.monitoring.record_event_duration_secs(
      '/orbax/emergency/checkpoint/read/transfer_global_shard_duration_secs',
      transfer_elapsed_s,
  )

  return resharded_state



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configurable options for customizing checkpointing behavior."""

from __future__ import annotations

import dataclasses
from typing import Any
from typing import Protocol

import numpy as np
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


@dataclasses.dataclass(frozen=True, kw_only=True)
class PyTreeOptions:
  """Options for PyTree checkpointing.

  Attributes:
    use_ocdbt: Whether to use OCDBT for saving.
    use_zarr3: Whether to use Zarr3 for saving.
    save_concurrent_bytes: The maximum number of bytes to save concurrently.
    restore_concurrent_bytes: The maximum number of bytes to restore
      concurrently.
    ocdbt_target_data_file_size: Specifies the target size (in bytes) of each
      OCDBT data file.  It only applies when OCDBT is enabled and Zarr3 must be
      turned on.  If left unspecified, default size is 2GB. A value of 0
      indicates no maximum file size limit. For best results, ensure
      chunk_byte_size is smaller than this value. For more details, refer to
      https://google.github.io/tensorstore/kvstore/ocdbt/index.html#json-kvstore/ocdbt.target_data_file_size
    enable_pinned_host_transfer: If False, disables transfer to pinned host when
      copying from device to host, regardless of the presence of pinned host
      memory.
    partial_load: If the tree structure omits some keys relative to the
      checkpoint, the omitted keys will not be loaded.
    create_array_storage_options_fn: A function that is applied to each leaf of
      the input PyTree (via `jax.tree.map_with_path`) to create a
      `ArrayStorageOptions` object, which is used to customize saving behavior
      for individual leaves. See `ArrayStorageOptions` and
      `CreateArrayStorageOptionsFn` for more details.
  """

  use_ocdbt: bool = True
  use_zarr3: bool = True
  save_concurrent_bytes: int | None = None
  restore_concurrent_bytes: int | None = None
  ocdbt_target_data_file_size: int | None = None
  enable_pinned_host_transfer: bool = False
  partial_load: bool = False
  create_array_storage_options_fn: CreateArrayStorageOptionsFn | None = None


@dataclasses.dataclass
class ArrayStorageOptions:
  """Arguments used to customize array storage behavior for individual leaves.

  dtype:
    If provided, casts the parameter to the given dtype before saving.
    Note that the parameter must be compatible with the given type (e.g.
    jnp.bfloat16 is not compatible with np.ndarray).
  chunk_byte_size:
    This is an experimental feature that automatically chooses the largest chunk
    shape possible, while keeping the chunk byte size less than or equal to the
    specified chunk_byte_size. Both the write_chunk_shape and read_chunk_shape
    are automatically set to the chosen shape. This uses a greedy algorithm that
    prioritizes splitting the largest dimensions first.
  shard_axes: An optional list of axes that should be prioritized when
    sharding array for storage. If empty, storage sharding implementation will
    prioritize axes which are already sharded.
  """

  dtype: np.typing.DTypeLike | None = None
  chunk_byte_size: int | None = None
  shard_axes: tuple[int, ...] = tuple()


class CreateArrayStorageOptionsFn(Protocol):

  def __call__(
      self, key: tree_types.PyTreeKeyPath, value: Any
  ) -> ArrayStorageOptions:
    ...



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Composite key-value store for checkpointing."""

from collections.abc import Iterator, KeysView, Mapping, ValuesView
from typing import Any


class Composite(Mapping):
  """A key-value store similar to a dict.

  Usage examples::

    Composite(
        state=my_train_state,
        dataset=my_dataset,
        metadata=json_metadata,
    )

    Composite(
        **{
            'state': my_train_state,
            'dataset': my_dataset,
            'metadata': json_metadata,
          }
    )

    args = Composite(...)
    args.state
    args['state']
    'state' in args

    for key, value in args.items():
      ...
  """

  _items: Mapping[str, Any]

  def __init__(self, **items: Any):
    super().__setattr__('_items', items)

    reserved_keys = set(dir(self))

    for key, value in items.items():
      # Reserve and prevent users from setting keys that start with '__'. These
      # may be used later to define options.
      if key.startswith('__'):
        raise ValueError(f'Composite keys cannot start with "__". Got: {key}')
      if key not in reserved_keys:
        # We do not raise an error if the user specifies a key that matches an
        # existing attribute (like 'keys', 'values', 'items'). These can be
        # accessed through self[key], but not self.key.
        super().__setattr__(key, value)

  def __getitem__(self, key: str) -> Any:
    if key not in self._items:
      raise KeyError(
          f'Unknown key: {key}. Available keys: {self._items.keys()}'
      )
    return self._items[key]

  def __iter__(self) -> Iterator[Any]:
    return iter(self._items)

  def __len__(self) -> int:
    return len(self._items)

  def __getattr__(self, key: str):
    return self.__getitem__(key)

  def __setattr__(self, key: str, value: Any):
    raise ValueError('Composite is immutable after initialization.')

  def get(self, key: str, default=None) -> Any | None:
    try:
      return self.__getitem__(key)
    except KeyError:
      return default

  def __and__(self, other: 'Composite') -> 'Composite':
    if isinstance(other, dict):
      other = Composite(**other)
    common_keys = self._items.keys() & other._items.keys()
    return Composite(**{key: self._items[key] for key in common_keys})

  def __or__(self, other: 'Composite') -> 'Composite':
    if isinstance(other, dict):
      other = Composite(**other)
    return Composite(**{**self._items, **other._items})

  def __repr__(self):
    return f'Composite({repr(self._items)})'

  def keys(self) -> KeysView[str]:
    return self._items.keys()

  def values(self) -> ValuesView[Any]:
    return self._items.values()

  def __contains__(self, key: str) -> bool:
    return key in self._items



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines policies for when a checkpoint is saved."""

import dataclasses
import typing
from typing import Container, Protocol, Sequence


@dataclasses.dataclass(kw_only=True)
class StepInfo:
  """Relevant information about a checkpoint step."""
  step: int


@dataclasses.dataclass(kw_only=True)
class DecisionContext:
  """Additional properties for making a save decision."""

  is_saving_in_progress: bool
  reached_preemption: bool


@typing.runtime_checkable
class SaveDecisionPolicy(Protocol):
  """A policy that defines when to save a checkpoint.

  Implementations should return True from `should_save` when saving a checkpoint
  is desired at the given step.
  """

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    ...


@dataclasses.dataclass
class FixedIntervalPolicy(SaveDecisionPolicy):
  """Checkpoint at a fixed interval."""

  interval: int

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    del previous_steps
    del context
    return step.step % self.interval == 0


@dataclasses.dataclass
class SpecificStepsPolicy(SaveDecisionPolicy):
  """Checkpoint at specific steps."""

  steps: Container[int]

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    del previous_steps
    del context
    return step.step in self.steps


class ContinuousCheckpointingPolicy(SaveDecisionPolicy):
  """Checkpoint as often as possible, as long as a save is not ongoing."""

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    del step
    del previous_steps
    return not context.is_saving_in_progress


class PreemptionCheckpointingPolicy(SaveDecisionPolicy):
  """Save a checkpoint when a preemption is detected."""

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    del step
    del previous_steps
    return context.reached_preemption


class InitialSavePolicy(SaveDecisionPolicy):
  """Checkpoint as soon as possible if no checkpoints already exist."""

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    del step
    del context
    return not previous_steps


@dataclasses.dataclass
class AnySavePolicy(SaveDecisionPolicy):
  """Evaluates all policies and saves if any of them returns True.

  Each policy is evaluated in order, and if all are False, the final result is
  False. If at least one is True, the final result is True.
  """

  policies: Sequence[SaveDecisionPolicy]

  def should_save(
      self,
      step: StepInfo,
      previous_steps: Sequence[StepInfo],
      *,
      context: DecisionContext
  ) -> bool:
    return any(
        policy.should_save(step, previous_steps=previous_steps, context=context)
        for policy in self.policies
    )



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multislice utilities."""

import functools
from typing import Any, Optional, Set, Tuple, Union

from absl import logging
import jax
from jax import numpy as jnp
import numpy as np
from orbax.checkpoint._src.multihost import multihost

PyTree = Any

# When using broadcasting from single replica to others, 3 copies of the data
# are stored in memory.
MEMORY_FACTOR = 3


def process_slice_id(
    process_index: int,
    global_mesh: jax.sharding.Mesh,
    *,
    replica_axis_index: int = 0,
) -> int:
  """Returns the slice id that the process_index belongs to."""
  for slice_id in range(
      slice_count(global_mesh, replica_axis_index=replica_axis_index)
  ):
    device_slice = slice_devices(
        global_mesh, replica_id=slice_id, replica_axis_index=replica_axis_index
    )
    if process_index in multihost.unique_processes_from_devices(device_slice):
      return slice_id
  return -1


def _process_in_device_slice(
    process_index: int, device_slice: np.ndarray
) -> bool:
  return process_index in multihost.unique_processes_from_devices(device_slice)


def slice_devices(
    global_mesh: jax.sharding.Mesh,
    *,
    replica_id: int = 0,
    replica_axis_index: int = 0,
) -> np.ndarray:
  devices = global_mesh.devices
  if hasattr(jax.devices()[0], 'slice_index'):
    get_slice_id = np.vectorize(lambda x: x.slice_index)
    return devices[get_slice_id(devices) == replica_id]
  else:
    return np.take(
        global_mesh.devices,
        replica_id,
        axis=replica_axis_index,
    )


def slice_count(
    global_mesh: jax.sharding.Mesh, *, replica_axis_index: int = 0
) -> int:
  """Number of slices implied by the mesh's replica dimension."""
  if len(global_mesh.shape_tuple) == 1:
    return 1
  return global_mesh.devices.shape[replica_axis_index]


def local_slice_devices(
    global_mesh: jax.sharding.Mesh, *, replica_axis_index: int = 0
) -> np.ndarray:
  """Get devices in the host-local slice."""
  for replica_id in range(
      slice_count(global_mesh, replica_axis_index=replica_axis_index)
  ):
    if in_slice(
        multihost.process_index(),
        global_mesh,
        replica_id=replica_id,
        replica_axis_index=replica_axis_index,
    ):
      return slice_devices(
          global_mesh,
          replica_id=replica_id,
          replica_axis_index=replica_axis_index,
      )
  raise ValueError(
      f'process_index {multihost.process_index()} does not exist in provided'
      ' `global_mesh`'
  )


def primary_process_in_slice(
    global_mesh: jax.sharding.Mesh,
    *,
    replica_id: int = 0,
    replica_axis_index: int = 0,
) -> int:
  """Returns an arbitrary process in the requested slice to serve as primary."""
  device_slice = slice_devices(
      global_mesh,
      replica_axis_index=replica_axis_index,
      replica_id=replica_id,
  )
  processes = multihost.unique_processes_from_devices(device_slice)
  return next(iter(processes))


def in_slice(
    process_index: int,
    global_mesh: jax.sharding.Mesh,
    *,
    replica_id: int = 0,
    replica_axis_index: int = 0,
) -> bool:
  """Returns if the process belongs to the indicated slice ID."""
  return _process_in_device_slice(
      process_index,
      slice_devices(
          global_mesh,
          replica_id=replica_id,
          replica_axis_index=replica_axis_index,
      ),
  )


@functools.partial(jax.jit, static_argnums=0)
def fake_zero_data(sharding, x):
  x = jnp.zeros_like(x)
  return jax.lax.with_sharding_constraint(x, sharding)


def get_device_memory() -> int:
  """Returns HBM capacity of the device on which the code is running(in bytes)."""
  device = jax.devices()[0]
  if device.platform not in ('tpu', 'gpu'):
    raise ValueError('Only select TPU and GPU devices are supported.')
  hbm_memory = {
      'TPU v3': int(16e9),  # two cores per chip each with 16 GB HBM
      'TPU v4': int(32e9),  # one megacore per chip with 32 GB HBM
      'TPU v5 lite': int(16e9),  # one core per chip with 16 GB HBM
      'TPU v5': int(96e9),  # one megacore per chip with 96 GB HBM
      'TPU v6 lite': int(32e9),  # one core per chip with 32 GB HBM
      'NVIDIA H100': int(144e9),
      'NVIDIA H200': int(80e9),
  }
  memory = hbm_memory.get(device.device_kind, None)
  if memory is None:
    raise ValueError(
        f'get_device_memory is not supported for {device.device_kind}.'
    )
  return memory


def get_leaf_memory_per_device(arr: jax.Array) -> int:
  """Returns the memory usage of a sharded array per device (in bytes)."""
  shard = arr.addressable_shards[0]
  return shard.data.size * shard.data.itemsize


def tree_memory_per_device(tree: Tuple[PyTree, ...]) -> int:
  """Returns the memory usage of a PyTree on each device (in bytes)."""
  leaf_memory_per_device = jax.tree_util.tree_map(
      get_leaf_memory_per_device, tree
  )
  return jax.tree.reduce(lambda x, y: x + y, leaf_memory_per_device)


def get_available_memory(
    in_tree: Tuple[PyTree, ...], scaling_factor: float
) -> int:
  """Returns estimated available memory for broadcasting (in bytes).

  After computing the available memory, we scale it by a factor of 0.75 to
  account for the fact that the actual memory usage could be different than the
  estimated memory usage. This will help us to avoid OOM errors for edge cases.

  Args:
    in_tree: pytree that occupies the memory.
    scaling_factor: indicates the frunction of the estimated available memory to
      be used when broadcustind data.
  """
  if scaling_factor > 1:
    raise ValueError('scaling_factorshould be less than 1.')
  total_device_memory = get_device_memory()
  used_device_memory = tree_memory_per_device(in_tree)
  available_memory = total_device_memory - used_device_memory
  return int(available_memory * scaling_factor / MEMORY_FACTOR)


def broadcast_one_replica_to_all(
    in_tree: Tuple[PyTree, ...],
    global_mesh: jax.sharding.Mesh,
    replica_axis_index: int,
    is_source: bool,
    memory_limit_bytes: Optional[Union[int, None]] = None,
    memory_scaling_factor: Optional[float] = 0.75,
) -> Tuple[Tuple[PyTree, ...], int]:
  """One replica reads the data and broadcasts to others.

  Args:
    in_tree: pytree to be broadcast. Shardings should correspond to the origin
      replica.
    global_mesh: global mesh.
    replica_axis_index: axis index along which the data is replicated.
    is_source: indicates if the current host is in origin replica.
    memory_limit_bytes: memory limit for broadcasting in bytes.
    memory_scaling_factor: indicates the fraction of the estimated available
      memory to be used when broadcasting data.

  Returns:
     Tuple containing:
      - pytree with broadcasted data
      - number of broadcasts performed.
  """
  num_replicas = global_mesh.devices.shape[replica_axis_index]
  replica_axis_name = global_mesh.axis_names[replica_axis_index]

  if memory_limit_bytes is None:
    memory_limit_bytes = get_available_memory(in_tree, memory_scaling_factor)
    logging.info('Using available memory of %d bytes.', memory_limit_bytes)

  # Set replica_axis to be 0, regardless of its actual value.
  def globalize_single_replica_arrays(inp):
    sharding = inp.sharding
    if not isinstance(sharding, jax.sharding.NamedSharding):
      raise ValueError(
          'Must provide input arrays with NamedSharding. '
          f'Got {type(sharding)} instead.'
      )
    if not is_source:
      inp = fake_zero_data(sharding, inp)
    inp = jnp.expand_dims(inp, axis=0)
    in_spec = jax.sharding.PartitionSpec(
        replica_axis_name,
        *sharding.spec,
    )
    global_shape = (num_replicas,) + inp.shape[1:]
    global_sharding = jax.sharding.NamedSharding(global_mesh, in_spec)
    return jax.make_array_from_single_device_arrays(
        global_shape, global_sharding, [s.data for s in inp.addressable_shards]
    )

  tree_len = len(in_tree)
  start = 0
  out_tree = []
  num_broadcasts = 0
  while start < tree_len:
    subtree = []
    current_memory = 0
    end = start
    if tree_memory_per_device(in_tree[start]) > memory_limit_bytes:
      logging.warning(
          'in_tree leaf size exceeds memory limit for broadcasting. '
          'Leaf size: %d bytes. Allowed memory limit: %d bytes. Proceeding.',
          tree_memory_per_device(in_tree[start]),
          memory_limit_bytes,
      )
      subtree.append(in_tree[end])
      end += 1
    else:
      while end < tree_len and (
          current_memory + tree_memory_per_device(in_tree[end])
          <= memory_limit_bytes
      ):
        subtree.append(in_tree[end])
        current_memory += tree_memory_per_device(in_tree[end])
        end += 1
    subtree = tuple(subtree)
    num_broadcasts += 1
    out_sharding = jax.tree.map(
        lambda x: jax.sharding.NamedSharding(
            global_mesh, jax.sharding.PartitionSpec(*x.sharding.spec)
        ),
        subtree,
    )
    in_tree_sharded = jax.tree.map(globalize_single_replica_arrays, subtree)
    # Delete immediately to conserve memory.
    jax.tree.map(lambda x: x.delete(), subtree)

    out_subtree = jax.jit(
        lambda tree: jax.tree.map(functools.partial(jnp.sum, axis=0), tree),
        out_shardings=out_sharding,
    )(in_tree_sharded)
    out_tree.extend(out_subtree)
    jax.block_until_ready(out_subtree)
    start = end

  if is_source:
    logging.info('Total number of broadcasts: %d', num_broadcasts)
  return tuple(out_tree), num_broadcasts


def get_primary_replica_ids_and_pids(
    replica_axis_idx: int,
    mesh: jax.sharding.Mesh,
    primary_replica_id: int,
) -> Tuple[Set[int], Set[int]]:
  """Returns the primary replica ids and process ids."""
  replica_devices = slice_devices(
      mesh,
      replica_id=primary_replica_id,
      replica_axis_index=replica_axis_idx,
  ).flatten()
  ids = set([d.id for d in replica_devices])
  pids = set([d.process_index for d in replica_devices])
  return ids, pids



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for array creation for unit tests."""

from typing import cast

import jax
import numpy as np
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.arrays import abstract_arrays
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


def create_sharded_array(
    arr: np.ndarray, sharding: jax.sharding.Sharding
) -> jax.Array:
  sharding = cast(jax.sharding.NamedSharding, sharding)
  return test_utils.create_sharded_array(arr, sharding.mesh, sharding.spec)


def create_numpy_pytree(*, add: int = 0, include_scalars: bool = True):
  pytree = test_utils.setup_pytree(add=add)
  if include_scalars:
    pytree.update({'x': 4.5, 'y': 3})
  return pytree, jax.tree.map(as_abstract_type, pytree)


def create_sharded_pytree(
    *, add: int = 0, reverse_devices: bool = False, include_scalars: bool = True
) -> tuple[tree_types.PyTree, tree_types.PyTree]:
  """Creates a sharded PyTree from `create_numpy_pytree`.

  Args:
    add: The value to add to leaf arrays.
    reverse_devices: Whether to reverse the devices in the mesh.
    include_scalars: Whether to include scalars in the pytree.

  Returns:
    A tuple of (pytree, abstract_pytree).
  """
  devices = jax.devices()
  num_devices = len(devices)
  devices = (
      np.asarray(list(reversed(devices)))
      if reverse_devices
      else np.asarray(devices)
  )

  mesh_2d = jax.sharding.Mesh(
      devices.reshape((2, num_devices // 2)), ('x', 'y')
  )
  mesh_axes_2d = jax.sharding.PartitionSpec('x', 'y')
  mesh_1d = jax.sharding.Mesh(devices, ('x',))
  mesh_axes_1d = jax.sharding.PartitionSpec(
      'x',
  )
  mesh_0d = jax.sharding.Mesh(devices, ('x',))
  mesh_axes_0d = jax.sharding.PartitionSpec(
      None,
  )

  shardings = {
      'a': jax.sharding.NamedSharding(mesh_0d, mesh_axes_0d),
      'b': jax.sharding.NamedSharding(mesh_1d, mesh_axes_1d),
      'c': {
          'a': jax.sharding.NamedSharding(mesh_2d, mesh_axes_2d),
          'e': jax.sharding.NamedSharding(mesh_2d, mesh_axes_2d),
      },
  }
  if include_scalars:
    shardings.update({
        'x': jax.sharding.NamedSharding(mesh_0d, mesh_axes_0d),
        'y': jax.sharding.NamedSharding(mesh_0d, mesh_axes_0d),
    })
  pytree, _ = create_numpy_pytree(add=add, include_scalars=include_scalars)
  pytree = jax.tree.map(create_sharded_array, pytree, shardings)
  return pytree, jax.tree.map(as_abstract_type, pytree)


def as_abstract_type(value):
  if isinstance(value, jax.Array):
    return abstract_arrays.to_shape_dtype_struct(value)
  elif isinstance(value, np.ndarray):
    return np.empty_like(value)
  elif isinstance(value, int):
    return 0
  elif isinstance(value, float):
    return 0.0
  else:
    raise ValueError(f'Unsupported type: {type(value)}')



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Synchronous Checkpointer implementation."""

import time
from typing import Any, Iterable, Optional, Type

from absl import logging
from etils import epath
from etils import epy
import jax
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint import utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.checkpointers import abstract_checkpointer
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.handlers import checkpoint_handler
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import atomicity_defaults
from orbax.checkpoint._src.path import atomicity_types
from typing_extensions import Self  # for Python version < 3.11



CheckpointArgs = checkpoint_args.CheckpointArgs
register_with_handler = checkpoint_args.register_with_handler
get_legacy_handler_wrapper = (
    composite_checkpoint_handler.get_legacy_handler_wrapper
)
StepMetadata = checkpoint.StepMetadata


def construct_checkpoint_args(
    handler: checkpoint_handler.CheckpointHandler,
    for_save: bool,
    *args,
    **kwargs,
) -> checkpoint_args.CheckpointArgs:
  """Constructs `CheckpointArgs` for save or restore for the handler."""
  for arg in args:
    if isinstance(arg, checkpoint_args.CheckpointArgs):
      return arg
  for arg in kwargs.values():
    if isinstance(arg, checkpoint_args.CheckpointArgs):
      return arg
  jax.monitoring.record_event('/jax/orbax/deprecation/checkpointer_legacy_args')
  save_arg_cls, restore_arg_cls = checkpoint_args.get_registered_args_cls(
      handler
  )
  if for_save:
    return save_arg_cls(*args, **kwargs)
  else:
    return restore_arg_cls(*args, **kwargs)


class Checkpointer(
    abstract_checkpointer.AbstractCheckpointer, epy.ContextManager
):
  """A synchronous implementation of AbstractCheckpointer.

  This class saves synchronously to a given directory using an underlying
  `CheckpointHandler`. Atomicity of the operation is guaranteed.

  IMPORTANT: Async checkpointing can often be faster for saving. Strongly
  consider using `AsyncCheckpointer` instead.

  IMPORTANT: Remember that to save and restore a checkpoint, one should always
  use an `AbstractCheckpointer` coupled with a `CheckpointHandler`. The specific
  `CheckpointHandler` to use depends on the object being saved or restored.

  Basic example::

    ckptr = Checkpointer(StandardCheckpointHandler())
    args = ocp.args.StandardSave(item=pytree_of_arrays)
    ckptr.save(path, args=args)
    args = ocp.args.StandardRestore(item=abstract_pytree_target)
    ckptr.restore(path, args=args)

  Each handler includes `...SaveArgs` and `...RestoreArgs` classes that document
  what arguments are expected. When using `Checkpointer`, you can either use
  this dataclass directly, or you can provide the arguments in keyword form.

  For example::

    ckptr = Checkpointer(StandardCheckpointHandler())
    ckptr.save(path, state=pytree_of_arays)
    ckptr.restore(path, state=abstract_pytree_target)
  """

  def __init__(
      self,
      handler: checkpoint_handler.CheckpointHandler,
      *,
      multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
      file_options: options_lib.FileOptions = options_lib.FileOptions(),
      checkpoint_metadata_store: Optional[checkpoint.MetadataStore] = None,
      temporary_path_class: Optional[
          Type[atomicity_types.TemporaryPath]
      ] = None,
  ):
    if not checkpoint_args.has_registered_args(handler):
      logging.warning(
          'No registered CheckpointArgs found for handler type: %s',
          type(handler),
      )
      handler = get_legacy_handler_wrapper(handler)
    self._handler = handler
    self._primary_host = multiprocessing_options.primary_host
    self._active_processes = multiprocessing_options.active_processes
    self._barrier_sync_key_prefix = (
        multiprocessing_options.barrier_sync_key_prefix
    )
    self._multiprocessing_options = options_lib.MultiprocessingOptions(
        primary_host=self._primary_host,
        active_processes=self._active_processes,
        barrier_sync_key_prefix=self._barrier_sync_key_prefix,
    )
    self._file_options = file_options
    self._temporary_path_class = temporary_path_class

    # If not provided then use checkpoint_metadata_store with blocking writes.
    self._metadata_store = (
        checkpoint_metadata_store
        or checkpoint.metadata_store(enable_write=True, blocking_write=True)
    )
    if not self._metadata_store.is_blocking_writer():
      raise ValueError('Checkpoint metadata store must be blocking writer.')

    jax.monitoring.record_event('/jax/orbax/checkpointer/init')

  def get_temporary_path(
      self, directory: epath.Path
  ) -> atomicity_types.TemporaryPath:
    temporary_path_class = (
        self._temporary_path_class
        or atomicity_defaults.get_default_temporary_path_class(directory)
    )
    tmpdir = temporary_path_class.from_final(
        directory,
        checkpoint_metadata_store=self._metadata_store,
        multiprocessing_options=self._multiprocessing_options,
        file_options=self._file_options,
    )
    return tmpdir

  async def create_temporary_path(
      self, temporary_path: atomicity_types.TemporaryPath
  ):
    await atomicity.create_all(
        [temporary_path],
        multiprocessing_options=self._multiprocessing_options,
    )

  def synchronize_next_awaitable_signal_operation_id(self):
    synchronization.HandlerAwaitableSignalOperationIdGenerator.next_operation_id()

    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'next_awaitable_signal_operation_id:sync',
            prefix=self._barrier_sync_key_prefix,
        ),
        timeout=multihost.DIRECTORY_CREATION_TIMEOUT,
        processes=self._active_processes,
    )

  def save(
      self,
      directory: epath.PathLike,
      *args,
      force: bool = False,
      custom_metadata: dict[str, Any] | None = None,
      **kwargs,
  ):
    """Saves the given item to the provided directory.

    Delegates to the underlying CheckpointHandler. Ensures save operation
    atomicity.

    This method should be called by all hosts - process synchronization and
    actions that need to be performed on only one host are managed internally.

    Args:
      directory: a path to which to save.
      *args: additional args to provide to the CheckpointHandler's save method.
      force: if True, allows overwriting an existing directory. May add overhead
        due to the need to delete any existing files.
      custom_metadata: a dictionary of custom metadata to be written to the
        checkpoint directory via StepMetadata.
      **kwargs: additional keyword args to provide to the CheckpointHandler's
        save method.

    Raises:
      ValueError if the provided directory already exists.
    """
    checkpoint_start_time = time.time()
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'Checkpointer:save_start',
            prefix=self._barrier_sync_key_prefix,
        ),
        processes=self._active_processes,
        record_event_name=(
            '/jax/checkpoint/write/checkpoint_start_sync_duration_secs'
        ),
    )
    directory = epath.Path(directory)

    jax.monitoring.record_event('/jax/orbax/write/start')
    logging.info(
        '[process=%s] Started saving checkpoint to %s.',
        multihost.process_index(),
        directory,
    )
    self.synchronize_next_awaitable_signal_operation_id()

    if directory.exists():
      if force:
        if utils.is_primary_host(self._primary_host):
          logging.info('Specified `force`: removing existing directory.')
          directory.rmtree()  # Post-sync handled by create_tmp_directory.
      else:
        raise ValueError(f'Destination {directory} already exists.')
    ckpt_args = construct_checkpoint_args(self._handler, True, *args, **kwargs)
    tmpdir = self.get_temporary_path(directory)
    # tmpdir creation also does an initial StepMetadata save.
    asyncio_utils.run_sync(self.create_temporary_path(tmpdir))
    self._handler.save(tmpdir.get(), args=ckpt_args)
    if utils.is_primary_host(self._primary_host):
      # Update StepMetadata after the handler save is complete. (blocking write)
      self._save_step_metadata(tmpdir.get(), custom_metadata=custom_metadata)
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'Checkpointer:save',
            prefix=self._barrier_sync_key_prefix,
        ),
        processes=self._active_processes,
    )

    # Ensure save operation atomicity and record time saved by checkpoint.
    if utils.is_primary_host(self._primary_host):
      # finalize does a final StepMetadata update.
      self._handler.finalize(tmpdir.get())
      atomicity.on_commit_callback(
          tmpdir,
          checkpoint_start_time=checkpoint_start_time,
      )
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'Checkpointer:finalize',
            prefix=self._barrier_sync_key_prefix,
            # suffix=directory.name,
        ),
        processes=self._active_processes,
    )
    save_duration_secs = time.time() - checkpoint_start_time
    logging.info(
        'Finished synchronous save in %.2f seconds to %s',
        save_duration_secs,
        directory,
    )

  def restore(self, directory: epath.PathLike, *args, **kwargs) -> Any:
    """See superclass documentation."""
    restore_start_time = time.time()
    directory = epath.Path(directory)
    if not directory.exists():
      raise FileNotFoundError(f'Checkpoint at {directory} not found.')
    if not utils.is_checkpoint_finalized(directory):
      raise ValueError(f'Found incomplete checkpoint at {directory}.')
    logging.info('Restoring checkpoint from %s.', directory)
    ckpt_args = construct_checkpoint_args(self._handler, False, *args, **kwargs)
    restored = self._restore(directory, args=ckpt_args)
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'Checkpointer:restore',
            prefix=self._barrier_sync_key_prefix,
        ),
        processes=self._active_processes,
    )
    restore_duration_secs = time.time() - restore_start_time
    logging.info(
        'Finished restoring checkpoint in %.2f seconds from %s.',
        restore_duration_secs,
        directory,
    )
    return restored

  def _restore(
      self, directory: epath.PathLike, args: checkpoint_args.CheckpointArgs
  ) -> Any:
    return self._handler.restore(directory, args=args)

  def metadata(self, directory: epath.PathLike) -> StepMetadata | Any | None:
    """See superclass documentation."""
    directory = epath.Path(directory)
    return self._handler.metadata(directory)

  def _save_step_metadata(
      self, directory: epath.Path, custom_metadata: dict[str, Any] | None
  ):
    """Saves StepMetadata to the checkpoint directory."""
    update_dict = {
        'custom_metadata': custom_metadata,
    }
    if isinstance(
        self._handler, composite_checkpoint_handler.CompositeCheckpointHandler
    ):
      try:
        # get item_handlers from handler
        partial_metadata: StepMetadata = (
            self._handler.metadata_from_temporary_paths(directory)
        )
      except (FileNotFoundError, NotImplementedError, ValueError, TypeError):
        logging.warning(
            'Failed to get per-item metadata from directory %s. Handler types '
            'will not be saved.',
            directory,
        )
      else:
        update_dict['item_handlers'] = partial_metadata.item_handlers
    else:
      try:
        item_handler = self._handler.typestr()
      except (NotImplementedError, AttributeError):
        logging.warning(
            'Failed to get item handler typestr from directory %s. Backup '
            'handler type will be saved.',
            directory,
        )
        item_handler = (
            f'{self._handler.__module__}.{self._handler.__class__.__qualname__}'
        )
      update_dict['item_handlers'] = item_handler
    self._metadata_store.update(
        file_path=checkpoint.step_metadata_file_path(directory),
        **step_metadata_serialization.serialize_for_update(**update_dict),
    )

  def close(self):
    """Closes the underlying CheckpointHandler."""
    self._handler.close()
    self._metadata_store.close()

  @property
  def handler(self) -> checkpoint_handler.CheckpointHandler:
    return self._handler

  def __contextmanager__(
      self,
  ) -> Iterable[Self]:
    try:
      yield self
    finally:
      self.close()



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A global registry of checkpoint handler types."""

from typing import Type

from absl import logging
from orbax.checkpoint._src.handlers import checkpoint_handler


CheckpointHandler = checkpoint_handler.CheckpointHandler


class HandlerTypeRegistry:
  """A registry mapping handler type strings to handler types."""

  def __init__(self):
    self._registry = {}

  def add(self, handler_type: Type[CheckpointHandler]) -> None:
    """Adds an entry to the registry."""
    try:
      handler_typestr = handler_type.typestr()
    except AttributeError:
      handler_typestr = f'{handler_type.__module__}.{handler_type.__qualname__}'
      logging.info(
          'Handler class %s does not have a typestr method. '
          'Using the default typestr value "%s" instead.',
          handler_type,
          handler_typestr,
      )
    if handler_typestr in self._registry:
      previous_handler_type = self._registry[handler_typestr]
      # On Colab/notebook, it's very common to reload modules when iterating
      # on the code (e.g. with `importlib.reload` or by re-executing cells).
      # This re-creates and re-registers a new handler class everytime.
      has_same_fully_qualified_name = (
          previous_handler_type.__module__ == handler_type.__module__
          and previous_handler_type.__qualname__ == handler_type.__qualname__
      )
      # If the fully-qualified-name changes then raise error.
      if not has_same_fully_qualified_name:
        raise ValueError(
            f'Handler type string "{handler_typestr}" already exists in the '
            f'registry with type {previous_handler_type}. '
            f'Cannot add type {handler_type}.'
        )
      # If both fully-qualified-name and type ref are the same then skip.
      if previous_handler_type == handler_type:
        logging.info(
            'Handler "%s" already exists in the registry with associated type'
            ' %s. Skipping registration.',
            handler_typestr,
            handler_type,
        )
        return
      # If fully-qualified-name is the same but type ref has changed then
      # it is okay to overwrite registry with the new type ref.
      logging.info(
          'Handler "%s" already exists in the registry with associated type'
          ' %s. Overwriting it as the module was recreated (likely from '
          'Colab reload',
          handler_typestr,
          handler_type,
      )
    self._registry[handler_typestr] = handler_type

  def get(
      self,
      handler_typestr: str,
  ) -> Type[CheckpointHandler]:
    """Gets an entry from the registry."""
    if handler_typestr not in self._registry:
      raise KeyError(
          f'Handler type string "{handler_typestr}" not found in the registry.'
      )
    return self._registry[handler_typestr]


_GLOBAL_HANDLER_TYPE_REGISTRY = HandlerTypeRegistry()


def register_handler_type(handler_cls):
  """Registers a checkpoint handler type in the global registry.

  The registry is keyed by the handler's typestr. If the handler does not
  provide a typestr, the default typestr is resolved from the handler's
  module and class name.

  Args:
    handler_cls: The checkpoint handler class to register.

  Returns:
    The registered checkpoint handler class.
  """
  _GLOBAL_HANDLER_TYPE_REGISTRY.add(handler_cls)
  return handler_cls


def get_handler_type(handler_typestr: str) -> Type[CheckpointHandler]:
  return _GLOBAL_HANDLER_TYPE_REGISTRY.get(handler_typestr)



# Copyright 2024 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Storage for PyTree checkpoint metadata with rich types."""

from __future__ import annotations

import collections
import functools
from typing import Any, Iterable, Mapping, Sequence, Type, TypeAlias

import jax
from orbax.checkpoint._src.metadata import pytree_metadata_options as pytree_metadata_options_lib
from orbax.checkpoint._src.metadata import value_metadata_entry
from orbax.checkpoint._src.tree import utils as tree_utils
import simplejson

PyTree: TypeAlias = Any


@functools.lru_cache()
def _new_namedtuple_type(
    module_name: str,
    class_name: str,
    fields: Sequence[str],
) -> Type[tuple[Any, ...]]:
  """Returns a namedtuple type created in the current module.

  NOTE: `module_name` and `class_name` are concatenated to create a unique
  class name to avoid name collisions.

  Args:
    module_name: Module name of original namedtuple saved in metadata.
    class_name: Class name of original namedtuple saved in metadata.
    fields: The fields of the namedtuple.
  """
  # TODO: b/365169723 - Return concrete NamedTuple if available in given module.
  arity = len(fields)
  unique_class_name = f'{module_name}_{class_name}_{arity}'
  # Valid class name must not contain dots.
  valid_class_name = unique_class_name.replace('.', '_')
  return collections.namedtuple(valid_class_name, fields)


def _create_namedtuple(
    *, module_name: str, class_name: str, attrs: Iterable[tuple[str, Any]]
) -> tuple[Any, ...]:
  """Returns a namedtuple instance with the given attributes and values, `attrs`.

  The namedtuple type is created in the current module on the fly using the
  given `module_name` and `class_name`. The two names are combined to create a
  unique class name to avoid name collisions. See `_new_namedtuple_type()` for
  more details.

  Args:
    module_name: Module name of original namedtuple saved in metadata.
    class_name: Class name of original namedtuple saved in metadata.
    attrs: The attributes of the namedtuple.
  """
  ks, vs = [*zip(*attrs)] or ((), ())
  result = _new_namedtuple_type(module_name, class_name, ks)(*vs)
  return result


def _module_and_class_name(cls) -> tuple[str, str]:
  """Returns the module and class name of the given class instance."""
  return cls.__module__, cls.__qualname__


_VALUE_METADATA_ENTRY_CLAZZ = 'ValueMetadataEntry'
_VALUE_METADATA_ENTRY_MODULE_AND_CLASS = _module_and_class_name(
    value_metadata_entry.ValueMetadataEntry
)


def _value_metadata_tree_for_json_dumps(obj: Any) -> Any:
  """Callback for `simplejson.dumps` to convert a PyTree to JSON object."""
  # Handle ValueMetadataEntry instances.
  if tree_utils.is_empty_or_leaf(obj):
    if (
        _module_and_class_name(obj.__class__)
        == _VALUE_METADATA_ENTRY_MODULE_AND_CLASS
    ):
      return dict(
          category='custom',
          clazz=_VALUE_METADATA_ENTRY_CLAZZ,
          data=obj.to_json(),
      )
    raise ValueError(
        f'Expected ValueMetadataEntry, got metadata pytree leaf: {obj}'
    )

  # Check namedtuple first and then tuple.
  if tree_utils.isinstance_of_namedtuple(obj):
    module_name, class_name = _module_and_class_name(obj.__class__)
    return dict(
        category='namedtuple',
        module=module_name,
        clazz=class_name,
        entries=[
            dict(key=k, value=_value_metadata_tree_for_json_dumps(v))
            for k, v in zip(obj._fields, obj)
        ],
    )
  # Check namedtuple first and then tuple.
  if isinstance(obj, tuple):
    return dict(
        category='custom',
        clazz='tuple',
        entries=[_value_metadata_tree_for_json_dumps(e) for e in obj],
    )

  if isinstance(obj, Mapping):
    return {k: _value_metadata_tree_for_json_dumps(v) for k, v in obj.items()}

  if isinstance(obj, list):
    return [_value_metadata_tree_for_json_dumps(e) for e in obj]

  # Handle objects that are registered as Jax container nodes.
  key_leafs, _ = jax.tree_util.tree_flatten_with_path(
      obj,
      is_leaf=lambda x: x is not obj,  # flatten just one level.
  )
  return {
      tree_utils.get_key_name(keypath[0]): _value_metadata_tree_for_json_dumps(
          leaf
      )
      for keypath, leaf in key_leafs
  }


def _value_metadata_tree_for_json_loads(obj):
  """Callback for `simplejson.loads` to convert JSON object to a PyTree."""
  if not isinstance(obj, Mapping):
    return obj

  if 'category' in obj:
    if obj['category'] == 'custom':
      if obj['clazz'] == _VALUE_METADATA_ENTRY_CLAZZ:
        return value_metadata_entry.ValueMetadataEntry.from_json(
            obj['data'],
            pytree_metadata_options_lib.PyTreeMetadataOptions(
                support_rich_types=True,  # Always in rich types mode.
            ),
        )
      if obj['clazz'] == 'tuple':
        return tuple(
            [(_value_metadata_tree_for_json_loads(v)) for v in obj['entries']]
        )
      raise ValueError(
          f'Unsupported "custom" object in JSON deserialization: {obj}'
      )

    if obj['category'] == 'namedtuple':
      return _create_namedtuple(
          module_name=obj['module'],
          class_name=obj['clazz'],
          attrs=[
              (
                  e['key'],
                  _value_metadata_tree_for_json_loads(e['value']),
              )
              for e in obj['entries']
          ],
      )

  return {k: _value_metadata_tree_for_json_loads(v) for k, v in obj.items()}


def value_metadata_tree_to_json_str(tree: PyTree) -> str:
  """Returns a JSON string representation of the given PyTree.

  Sample JSON::
  ```
  '{
    "mu_nu": {
      "category": "namedtuple",
      "module": "orbax.checkpoint._src.testing.test_tree_utils",
      "clazz": "MuNu",
      "entries": [
        {
          "key": "mu",
          "value": {
            "category": "custom",
            "clazz": "ValueMetadataEntry",
            "data": {
              "value_type": "jax.Array",
              "skip_deserialize": false
            }
          }
        },
        {
          "key": "nu",
          "value": {
            "category": "custom",
            "clazz": "ValueMetadataEntry",
            "data": {
              "value_type": "np.ndarray",
              "skip_deserialize": false
            }
          }
        }
      ]
    },
    "my_tuple": {
      "category": "custom",
      "clazz": "tuple",
      "entries": [
          {
            "category": "custom",
            "clazz": "ValueMetadataEntry",
            "data": {
              "value_type": "np.ndarray",
              "skip_deserialize": false
            }
          }
      ]
    }
  }'
  ```

  Args:
    tree: A PyTree to be converted to JSON string.
  """
  return simplejson.dumps(
      tree,
      default=_value_metadata_tree_for_json_dumps,
      tuple_as_array=False,  # Must be False to preserve tuples.
      namedtuple_as_object=False,  # Must be False to preserve namedtuples.
  )


def value_metadata_tree_from_json_str(json_str: str) -> PyTree:
  """Returns a PyTree from the given JSON string."""
  return simplejson.loads(
      json_str,
      object_hook=_value_metadata_tree_for_json_loads,
  )
